"""Storm analysis utilities."""

import datetime
import logging
import os

import numpy as np
import pandas as pd

from stormhub.utils import StacPathManager


class StormAnalyzer:
    """Analyze storm events."""

    def __init__(self, csv_path: str, threshold: float, duration_hours: int):
        self.csv_path = csv_path
        self.threshold = threshold
        self.duration_hours = duration_hours
        self.metrics_df = self._load_and_filter_data()
        if self.metrics_df.shape[0] == 0:
            raise ValueError("No storm events meet the threshold. Please adjust the threshold and try again.")
        self.filter = self._initialize_filter()

    def _load_and_filter_data(self) -> pd.DataFrame:
        """Load and filter storm data to above mean threshold."""
        df = pd.read_csv(self.csv_path)
        df["storm_date"] = pd.to_datetime(df["storm_date"])
        return df[df["mean"] >= self.threshold]

    def _initialize_filter(self):
        """Initialize storm filter."""
        if self.metrics_df["storm_date"].isnull().all():
            raise ValueError(
                "All storm_date values are NaT. Please check the input data and ensure events meet threshold."
            )

        start = self.metrics_df["storm_date"].min()
        end = self.metrics_df["storm_date"].max()

        if pd.isna(start) or pd.isna(end):
            raise ValueError("Start or end date is NaT. Please check the input data and ensure events meet threshold.")

        return StormFilter(start, end, datetime.timedelta(hours=1))

    def rank_and_filter_storms(self, buffer_hours: int = 24) -> pd.DataFrame:
        """Rank and filter storms.

        buffer_hours represent the time between potential storm events.
        TODO: Investigate appropriate values for this and verify functionality.
        """
        storm_records = []
        overlapping_overall_rank = 1
        non_overlapping_rank = 1
        year_rank_dict = {}
        year_rank = 1

        sorted_metrics_df = self.metrics_df.sort_values(by="mean", ascending=False)

        for _, row in sorted_metrics_df.iterrows():
            start_dt = row["storm_date"]
            end_dt = start_dt + datetime.timedelta(hours=self.duration_hours + buffer_hours)
            year = start_dt.year

            overlapping_year_rank, non_overlapping_year_rank = year_rank_dict.get(year, (1, 1))

            record = {
                "storm_date": start_dt.strftime("%Y-%m-%dT%H"),
                "mean": row["mean"],
                "min": row["min"],
                "max": row["max"],
                "overlapping_overall_rank": overlapping_overall_rank,
                "year_rank": year_rank,
            }
            overlapping_overall_rank += 1
            year_rank += 1

            if self.filter.try_block_period(start_dt, end_dt):
                record["non_overlapping_rank"] = non_overlapping_rank
                record["non_overlapping_year_rank"] = non_overlapping_year_rank
                non_overlapping_rank += 1
                non_overlapping_year_rank += 1
            else:
                record["non_overlapping_rank"] = -1
                record["non_overlapping_year_rank"] = -1

            year_rank_dict[year] = (overlapping_year_rank, non_overlapping_year_rank)
            storm_records.append(record)

        df = pd.DataFrame(storm_records)
        df.drop(columns=["overlapping_overall_rank", "year_rank"], inplace=True)
        df = df[df["non_overlapping_rank"] > 0]
        df.rename(
            columns={"non_overlapping_year_rank": "annual_rank", "non_overlapping_rank": "por_rank"}, inplace=True
        )

        return df.sort_values(by="por_rank").reset_index(drop=True)

    def _rank_storm_ids(self) -> list[str]:
        """Rank storm ids."""
        sorted_indices = np.argsort(self.metrics_df["mean"].values)[::-1]
        return self.metrics_df["storm_date"].dt.strftime("%Y-%m-%dT%H").iloc[sorted_indices].tolist()

    def rank_and_save(self, collection_id: str, spm: StacPathManager) -> pd.DataFrame:
        """Rank storms and save to csv."""
        output_file = os.path.join(spm.collection_dir(collection_id), "ranked-storms.csv")
        ranked_df = self.rank_and_filter_storms()
        ranked_df.to_csv(output_file, index=False)
        ranked_df["storm_date"] = pd.to_datetime(ranked_df["storm_date"])
        logging.info("Saved ranked storm data to %s", output_file)
        return ranked_df


class StormFilter:
    """Filter storm events from a series of cumulative grid statistics."""

    def __init__(
        self,
        start: datetime.datetime,
        end: datetime.datetime,
        interval: datetime.timedelta,
    ) -> None:
        self.start = start
        self.end = end
        self.interval = interval
        self.datetime_array = self._generate_datetime_array()

    def available_dates(self) -> np.ndarray:
        """Get available dates."""
        return self.datetime_array.compressed()

    def _generate_datetime_array(self) -> np.ma.MaskedArray:
        """Generate datetime array."""
        dt_list = np.arange(self.start, self.end, self.interval).astype("datetime64")
        return np.ma.array(data=dt_list, dtype=np.datetime64)

    def block_period(self, start: np.datetime64, end: np.datetime64) -> None:
        """Block a period of time."""
        mask = (self.datetime_array >= start) & (self.datetime_array < end)
        self.datetime_array[mask] = np.ma.masked

    def try_block_period(self, start: datetime.datetime, end: datetime.datetime) -> bool:
        """Try to block a period of time."""
        start_np = np.datetime64(start)
        end_np = np.datetime64(end)
        if start_np in self.available_dates() and end_np in self.available_dates():
            self.block_period(start_np, end_np)
            return True
        return False



"""Create a simple HTTP server for viewing local STAC objects."""

import os
import sys
import webbrowser
from http.server import HTTPServer, SimpleHTTPRequestHandler
from socketserver import ThreadingMixIn


class CORSRequestHandler(SimpleHTTPRequestHandler):
    """Handle CORS requests."""

    def end_headers(self):
        """Add CORS headers to the response."""
        self.send_header("Access-Control-Allow-Origin", "*")
        self.send_header("Access-Control-Allow-Methods", "GET, OPTIONS")
        self.send_header("Access-Control-Allow-Headers", "x-api-key, Content-Type")
        SimpleHTTPRequestHandler.end_headers(self)

    def list_directory(self, path):
        """List the contents of a directory."""
        try:
            list = os.listdir(path)
        except OSError:
            self.send_error(404, "No permission to list directory")
            return None
        list.sort(key=lambda a: a.lower())
        r = []
        displaypath = self.path
        r.append(f"<!DOCTYPE html><html><head><title>Directory listing for {displaypath}</title></head>")
        r.append(
            '<div style="text-align: right; margin: 20px;"><button style="background-color: #2986cc; color: \
             white; padding: 10px 20px; border: none; border-radius: 5px; font-size: 16px; cursor: pointer;" \
             onclick="stopServer()">🛑 Stop Server</button></div>'
        )
        r.append(
            f'<h2><a href="https://radiantearth.github.io/stac-browser/#/external/http://localhost:{self.server.server_port}/catalog.json" target="_blank">STAC Viewer</a></h2>'
        )
        r.append(f"<body><h4>Local directory listing for {displaypath}</h4>")

        r.append(
            """
            <script>
                function stopServer() {
                    fetch('/shutdown', { method: 'POST' })
                    .then(response => response.text())
                    .then(data => {
                        alert(data);
                        window.close();
                    })
                    .catch(error => alert('Failed to stop server'));
                }
            </script>
        """
        )

        r.append("<hr><ul>")
        for name in list:
            fullname = os.path.join(path, name)
            displayname = name
            linkname = name
            if os.path.isdir(fullname):
                displayname = name + "/"
                linkname = name + "/"
            r.append(f'<li><a href="{linkname}">{displayname}</a></li>')
        r.append("</ul><hr>")
        r.append("</body></html>")
        encoded = "\n".join(r).encode("utf-8", "surrogateescape")
        self.send_response(200)
        self.send_header("Content-Type", "text/html; charset=utf-8")
        self.send_header("Content-Length", str(len(encoded)))
        self.end_headers()
        self.wfile.write(encoded)
        return None

    def do_POST(self):
        """Handle POST requests."""
        if self.path == "/shutdown":
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(b"Server is shutting down...")
            print("Shutting down the server...")

            def shutdown():
                self.server.shutdown()

            import threading

            threading.Thread(target=shutdown).start()


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """Handle requests in a separate thread."""


def main(mkdir: bool = True):
    """Start a simple HTTP server."""
    if len(sys.argv) < 2:
        print("Usage: python server.py <directory_to_serve> [host] [port]")
        sys.exit(1)

    local_dir = sys.argv[1]
    if not os.path.isdir(local_dir):
        if mkdir:
            os.makedirs(local_dir)
        else:
            print(
                f"The specified directory '{local_dir}' does not exist or is not a directory., set mkdir=True to create it."
            )
            sys.exit(1)

    os.chdir(local_dir)

    host = sys.argv[2] if len(sys.argv) > 2 else "0.0.0.0"
    port = int(sys.argv[3]) if len(sys.argv) > 3 else 5000

    print(f"Serving '{local_dir}' on {host}:{port}")
    httpd = ThreadedHTTPServer((host, port), CORSRequestHandler)

    url = f"http://localhost:{port}/"
    webbrowser.open(url)

    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        print("\nShutting down the server.")
        httpd.server_close()


if __name__ == "__main__":
    main()



"""Script to validate dss files against expected lengths for each dss path Part C variable."""

import boto3
import os
import logging
from hecdss import HecDss
from dotenv import load_dotenv

# log to a file
logging.basicConfig(filename="dss_validation.log", level=logging.INFO, format="%(message)s")

load_dotenv()
s3 = boto3.client("s3")

EXPECTED_LENGTHS = {"PRECIPITATION": 72, "TEMPERATURE": 864}  # Expected number of part C dss paths.


def list_dss_files(bucket_name, prefix):
    """
    List all .dss files located directly under the specified S3 bucket and prefix.

    Yields
    ------
        str: S3 key for each .dss file found.
    """
    response = s3.list_objects_v2(Bucket=bucket_name, Prefix=prefix)
    for obj in response.get("Contents", []):
        key = obj["Key"]
        if key.endswith(".dss") and "/" not in key[len(prefix) :]:
            yield key


def check_dss_file(s3_key, bucket_name, download_dir):
    """Download and validate given dss s3 key."""
    filename = os.path.join(download_dir, os.path.basename(s3_key))
    s3.download_file(bucket_name, s3_key, filename)

    results = {"PRECIPITATION": 0, "TEMPERATURE": 0}

    try:
        with HecDss(filename) as dss:
            catalog = dss.get_catalog()
            for path_obj in catalog:
                path = str(path_obj)
                parts = path.strip("/").split("/")
                part_c = parts[2].upper()
                if part_c in results:
                    results[part_c] += 1
    except Exception as e:
        logging.error(f"ERROR opening {s3_key} - {str(e)}")
        return

    if results == EXPECTED_LENGTHS:
        logging.info(f"PASS - {s3_key} - {results}")
    else:
        logging.info(f"FAIL - {s3_key} - {results}")


def main(bucket_name, prefix, download_dir):
    """Check dss files."""
    os.makedirs(download_dir, exist_ok=True)

    for dss_key in list_dss_files(bucket_name, prefix):
        try:
            check_dss_file(dss_key, bucket_name, download_dir)
        except Exception as e:
            logging.error(f"ERROR - {dss_key} - {str(e)}")


if __name__ == "__main__":
    bucket_name = "trinity-pilot"
    prefix = "stac/prod-support/conformance/storm-catalog/storms/"
    download_dir = "dss_downloads"

    main(bucket_name, prefix, download_dir)



"""Module for processing and creating stormhub STAC objects."""

import json
import logging
import os
import traceback
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
from datetime import datetime, timedelta
from typing import Any, List, Union

import pandas as pd
import pystac
from pystac import Asset, Collection, Item, Link, MediaType
from shapely.geometry import mapping, shape, Point

from stormhub.hydro_domain import HydroDomain
from stormhub.logger import initialize_logger
from stormhub.met.analysis import StormAnalyzer
from stormhub.met.aorc.aorc import AORCItem, valid_spaces_item
from stormhub.utils import (
    STORMHUB_REF_LINK,
    StacPathManager,
    generate_date_range,
    validate_config,
)


def get_meteorological_season(month: int) -> str:
    """Return meteorological season given a month."""
    if month in [12, 1, 2]:
        return "Winter"
    elif month in [3, 4, 5]:
        return "Spring"
    elif month in [6, 7, 8]:
        return "Summer"
    else:
        return "Autumn"


class StormCollection(pystac.Collection):
    """Storm Collection class."""

    def __init__(self, collection_id: str, items: List[pystac.Item]):
        """
        Initialize a StormCollection instance.

        Args:
            collection_id (str): The ID of the collection.
            items (List[pystac.Item]): List of STAC items to include in the collection.
        """
        spatial_extents = [item.bbox for item in items if item.bbox]
        temporal_extents = [item.datetime for item in items if item.datetime is not None]

        collection_extent = pystac.Extent(
            spatial=pystac.SpatialExtent(
                bboxes=[
                    [
                        min(b[0] for b in spatial_extents),
                        min(b[1] for b in spatial_extents),
                        max(b[2] for b in spatial_extents),
                        max(b[3] for b in spatial_extents),
                    ]
                ]
            ),
            temporal=pystac.TemporalExtent(intervals=[[min(temporal_extents), max(temporal_extents)]]),
        )

        super().__init__(
            id=collection_id,
            description="STAC collection generated from storm items",
            extent=collection_extent,
        )

        for item in items:
            self.add_item_to_collection(item)

        self.add_link(STORMHUB_REF_LINK)

    @classmethod
    def from_collection(cls, collection: pystac.Collection) -> "StormCollection":
        """
        Create a StormCollection from an existing pystac.Collection.

        Args:
            collection (pystac.Collection): The existing STAC collection.

        Returns
        -------
            StormCollection: A new StormCollection instance.
        """
        items = list(collection.get_all_items())
        return cls(collection.id, items)

    def add_item_to_collection(self, item: Item, override: bool = False):
        """
        Add an item to the collection.

        Args:
            item (Item): The STAC item to add.
            override (bool): Whether to override an existing item with the same ID.
        """
        existing_ids = {item.id for item in self.get_all_items()}

        if item.id in existing_ids:
            if override:
                self.remove_item(item.id)
                self.add_item(item)
                logging.info("Overwriting (existing) item with ID '%s'.", item.id)
            else:
                logging.error(
                    "Item with ID '%s' already exists in the collection. Use `override=True` to overwrite.",
                    item.id,
                )

        else:
            self.add_item(item)
            logging.info("Added item with ID '%s' to the collection.", item.id)

    def add_summary_stats(self, spm: StacPathManager, property_name: str = "aorc:statistics", statistic: str = "mean"):
        """
        Add summary statistics to the collection.

        Args:
            spm (StacPathManager): The STAC path manager.
            property_name (str): The property name to summarize.
            statistic (str): The statistic to calculate (e.g., "mean").
        """
        values = []
        for item in self.get_all_items():
            if property_name in item.properties:
                values.append(item.properties[property_name].get(statistic))

        if values:
            min_value = min(values)
            max_value = max(values)
        else:
            min_value, max_value = None, None

        if "summaries" not in self.extra_fields:
            self.extra_fields["summaries"] = {}

        if min_value is not None and max_value is not None:
            self.extra_fields["summaries"][f"{property_name} precip (inches)"] = {
                "minimum": min_value,
                "maximum": max_value,
            }
        else:
            logging.warning("No values found for %s in collection: %s", property_name, self.id)
            self.extra_fields["summaries"][f"{property_name} precip (inches)"] = {
                "minimum": 0,
                "maximum": 0,
            }

        logging.info(
            "Summary statistics for %s: %s - %s saved at %s",
            property_name,
            min_value,
            max_value,
            spm.collection_file(self.id),
        )

        self.save_object(dest_href=spm.collection_file(self.id), include_self_link=False)

    def watershed_centroid_feature_collection(self, spm: StacPathManager):
        """
        Create a feature collection of watershed centroids from each storm event.

        Args:
            spm (StacPathManager): The STAC path manager.
        """
        features = []
        for item in self.get_all_items():
            geom = shape(item.geometry)
            if geom.is_empty:
                continue

            feature = {
                "type": "Feature",
                "geometry": mapping(geom),
                "properties": {
                    "id": item.id,
                    "storm_start_date": item.properties.get("start_datetime"),
                    "aorc:statistics": item.properties.get("aorc:statistics"),
                },
            }
            features.append(feature)

        feature_collection = {"type": "FeatureCollection", "features": features}

        output_geojson = spm.collection_asset(self.id, "transposed_watershed_centroids.geojson")
        with open(output_geojson, "w", encoding="utf-8") as f:
            json.dump(feature_collection, f, indent=4)

        self.add_asset(
            "transposed_watershed_centroids",
            Asset(
                href=spm.collection_asset(self.id, "transposed_watershed_centroids.geojson"),
                title="Transposed Watershed Centroids",
                description=f"Feature collection of Transposed Watershed Centroids from each storm event.",
                media_type=MediaType.GEOJSON,
                roles=["storm_summary"],
            ),
        )

        logging.info("FeatureCollection saved to %s", output_geojson)
        self.save_object(dest_href=spm.collection_file(self.id), include_self_link=False)

    def max_precip_feature_collection(self, spm: StacPathManager):
        """
        Create a feature collection of max precipitation locations from each storm event.

        Args:
            spm (StacPathManager): The STAC path manager.
        """
        features = []
        for item in self.get_all_items():
            loc = item.properties.get("aorc:max_precip_location")
            if not loc or "latitude" not in loc or "longitude" not in loc:
                continue

            storm_start = item.properties.get("start_datetime")
            storm_start_dt = datetime.strptime(storm_start, "%Y-%m-%dT%H:%M:%SZ")

            point = Point(loc["longitude"], loc["latitude"])
            feature = {
                "type": "Feature",
                "geometry": mapping(point),
                "properties": {
                    "id": item.id,
                    "storm_start_date": storm_start,
                    "aorc:statistics": item.properties.get("aorc:statistics"),
                    "season": get_meteorological_season(storm_start_dt.month),
                },
            }
            features.append(feature)

        feature_collection = {"type": "FeatureCollection", "features": features}

        output_geojson = spm.collection_asset(self.id, "max_precip_locations.geojson")
        with open(output_geojson, "w", encoding="utf-8") as f:
            json.dump(feature_collection, f, indent=4)

        self.add_asset(
            "max_precip_locations",
            Asset(
                href=spm.collection_asset(self.id, "max_precip_locations.geojson"),
                title="Max Precipitation Locations",
                description="Feature collection of max precipitation locations from each storm event.",
                media_type=MediaType.GEOJSON,
                roles=["storm_summary"],
            ),
        )

        logging.info("FeatureCollection saved to %s", output_geojson)
        self.save_object(dest_href=spm.collection_file(self.id), include_self_link=False)


class StormCatalog(pystac.Catalog):
    """
    Initialize a StormCatalog instance.

    Args:
        id (str): The ID of the catalog.
        watershed (HydroDomain): The watershed domain.
        transposition_region (HydroDomain): The transposition region domain.
        description (str): Description of the catalog.
        local_dir (str): Local directory for the catalog.
        valid_transposition_region (HydroDomain, optional): Valid transposition region domain.
        **kwargs (Any): Additional keyword arguments.
    """

    def __init__(
        self,
        catalog_id: str,
        watershed: HydroDomain,
        transposition_region: HydroDomain,
        description: str,
        local_dir: str,
        valid_transposition_region: HydroDomain = None,
        **kwargs: Any,
    ):
        super().__init__(id=catalog_id, description=description)
        self.catalog_type = pystac.CatalogType.SELF_CONTAINED
        self.local_dir = local_dir
        self.spm = StacPathManager(local_dir)
        self._watershed = self.add_hydro_domain(watershed)
        self._transposition_region = self.add_hydro_domain(transposition_region)
        if valid_transposition_region:
            self._valid_transposition_region = self.add_hydro_domain(valid_transposition_region)
        else:
            self._valid_transposition_region = None

        if "links" in kwargs:
            self.links = kwargs.get("links", [])
        self.set_self_href(self.spm.catalog_file)

    @classmethod
    def from_file(cls, file_path: str) -> "StormCatalog":
        """Create a StormCatalog from a file.

        Args:
            file_path (str): Path to the catalog file.

        Returns
        -------
            StormCatalog: A new StormCatalog instance.
        """
        catalog = pystac.Catalog.from_file(file_path)
        links = catalog.get_links()
        spm = StacPathManager(os.path.dirname(file_path))
        watershed = get_item_from_catalog_link(links, "Watershed", spm=spm)
        transposition_region = get_item_from_catalog_link(links, "Transposition Region", spm=spm)
        valid_transposition_region = get_item_from_catalog_link(links, "Valid Transposition Region", spm=spm)

        if not watershed or not transposition_region:
            raise ValueError("Both watershed and transposition region must be defined in the catalog.")

        return cls(
            catalog_id=catalog.id,
            watershed=watershed,
            transposition_region=transposition_region,
            description=catalog.description,
            local_dir=os.path.dirname(file_path),
            valid_transposition_region=valid_transposition_region,
            links=links,
        )

    @property
    def valid_transposition_region(self) -> Item:
        """
        Get the valid transposition region from the catalog links.

        Returns
        -------
            Item: The valid transposition region item.
        """
        # if self._valid_transposition_region is None:
        #     vtr_polygon = valid_spaces_item(self.watershed, self.transposition_region)
        #     vtr_id = f"{self.transposition_region.id}_valid"
        #     vtr = HydroDomain(
        #         id=vtr_id,
        #         geometry=vtr_polygon,
        #         hydro_domain_type="valid_transposition_region",
        #         description=f"Valid transposition region for {self.watershed.id} watershed",
        #         href=self.spm.catalog_item(vtr_id),
        #         title="Valid Transposition Region",
        #     )
        #     self.add_item(vtr)
        #     vtr.save_object(include_self_link=False)
        #     self._valid_transposition_region = vtr
        return get_item_from_catalog_link(self.links, "Valid Transposition Region", spm=self.spm)

    @property
    def transposition_region(self) -> Item:
        """
        Get the transposition region from the catalog links.

        Returns
        -------
            Item: The transposition region item.
        """
        return get_item_from_catalog_link(self.links, "Transposition Region", spm=self.spm)

    @property
    def watershed(self) -> Item:
        """
        Get the watershed from the catalog links.

        Returns
        -------
            Item: The watershed item.
        """
        return get_item_from_catalog_link(self.links, "Watershed", spm=self.spm)

    def sanitize_catalog_assets(self):
        """Force the asset paths in the catalog relative to root."""
        for collection in self.get_all_collections():
            for asset in collection.assets.values():
                if self.spm.collection_dir(collection.id).replace("\\", "/") in asset.href:
                    asset.href = asset.href.replace(self.spm.collection_dir(collection.id).replace("\\", "/"), ".")
                elif self.spm.catalog_dir.replace("\\", "/") in asset.href:
                    asset.href = asset.href.replace(self.spm.catalog_dir.replace("\\", "/"), "..")

            for item in collection.get_all_items():
                for asset in item.assets.values():
                    if self.spm.collection_item_dir(collection.id, item.id) in asset.href:
                        asset.href = asset.href.replace(self.spm.collection_item_dir(collection.id, item.id), ".")
                    elif self.spm.collection_dir(collection.id) in asset.href:
                        asset.href = asset.href.replace(self.spm.collection_dir(collection.id), ".")
                    elif self.spm.catalog_dir in asset.href:
                        asset.href = asset.href.replace(self.spm.catalog_dir, "..")
            collection.save()

    def add_hydro_domain(self, hydro_domain: Union[HydroDomain, Item]) -> str:
        """
        Add a hydro domain to the catalog.

        Args:
            hydro_domain (Union[HydroDomain, Item]): The hydro domain to add.

        Returns
        -------
            str: The ID of the added hydro domain.
        """
        if not isinstance(hydro_domain, (HydroDomain, Item)):
            raise ValueError(f"Expected a HydroDomain or an Item object not: {type(hydro_domain)}")
        try:
            title = hydro_domain.title
        except AttributeError:
            title = hydro_domain.id

        self.add_link(
            Link(
                rel="item",
                target=self.spm.catalog_asset(hydro_domain.id).replace(self.spm.catalog_dir, "."),
                title=title,
                media_type=pystac.MediaType.GEOJSON,
                extra_fields={
                    "Name": hydro_domain.id,
                    "Description": f"Input {hydro_domain.id} used to generate this catalog",
                },
            )
        )
        return hydro_domain.id

    def get_storm_collection(self, collection_id: str) -> StormCollection:
        """
        Get a storm collection from the catalog.

        Args:
            collection_id (str): The ID of the collection.

        Returns
        -------
            StormCollection: The storm collection.
        """
        collection = self.get_child(collection_id)
        if not collection:
            raise ValueError(f"Collection with ID '{collection_id}' not found in the catalog.")
        return StormCollection.from_collection(collection)

    def save_catalog(self):
        """Save the catalog and its collections."""
        for collection in self.get_all_collections():
            collection.save_object(dest_href=self.spm.collection_file(collection.id), include_self_link=False)
        self.sanitize_catalog_assets()
        self.save()

    def add_collection_to_catalog(self, collection: Collection, override: bool = False):
        """
        Add a collection to the catalog.

        Args:
            collection (Collection): The collection to add.
            override (bool): Whether to override an existing collection with the same ID.
        """
        existing_collections = {c.id for c in self.get_all_collections()}
        logging.info("Existing collection IDs: %s", existing_collections)

        if collection.id in existing_collections:
            if override:
                self.remove_child(collection.id)
                self.add_child(collection, title=collection.id)
                logging.info("Overwriting (existing) collection with ID '%s'.", collection.id)
            else:
                logging.error(
                    "Collection with ID '%s' already exists in the collection. Use `override=True` to overwrite.",
                    collection.id,
                )

        else:
            self.add_child(collection, title=collection.id)
            logging.info("Added collection with ID '%s' to the catalog.", collection.id)

    def new_collection_from_items(self, collection_id: str, items: List[Item]) -> StormCollection:
        """
        Create a new collection from a list of items.

        Args:
            collection_id (str): The ID of the new collection.
            items (List[Item]): List of items to include in the collection.

        Returns
        -------
            StormCollection: The new storm collection.
        """
        collection = StormCollection(collection_id, items)
        collection.add_asset(
            "valid-transposition-region",
            pystac.Asset(
                href=self.valid_transposition_region.self_href,
                title="Valid Transposition Region",
                description=f"Valid transposition region for {self.watershed.id} watershed",
                media_type=pystac.MediaType.GEOJSON,
                roles=["valid_transposition_region"],
            ),
        )

        collection.add_asset(
            "watershed",
            pystac.Asset(
                href=self.watershed.self_href,
                title="Watershed",
                description=f"{self.watershed.id} watershed",
                media_type=pystac.MediaType.GEOJSON,
                roles=["watershed"],
            ),
        )

        collection.save_object(dest_href=self.spm.collection_file(collection_id), include_self_link=False)
        self.add_collection_to_catalog(collection, override=True)
        self.sanitize_catalog_assets()
        return collection

    def sort_collection(self, collection_id: Collection, property_name: str):
        """
        Sort and save a STAC collection based on a specific property.

        Args:
            collection (Collection): The STAC collection to sort and save.
            property_name (str): The property name to sort by.
        """
        collection = self.get_storm_collection(collection_id)
        sorted_items = sorted(collection.get_all_items(), key=lambda item: item.properties.get(property_name))

        return StormCollection(collection.id, sorted_items)

    def add_rank_to_collection(
        self,
        collection_id: str,
        top_events: pd.DataFrame,
    ) -> StormCollection:
        """
        Create a new collection from a list of items.

        Args:
            collection_id (str): The ID of the new collection.
            items (List[Item]): List of items to include in the collection.

        Returns
        -------
            StormCollection: The new storm collection.
        """
        collection = self.get_storm_collection(collection_id)

        top_events.loc[:, "storm_id"] = top_events["storm_date"].apply(lambda x: f"{x.strftime('%Y-%m-%dT%H')}")
        for item in collection.get_all_items():
            matching_events = top_events[top_events["storm_id"] == item.id]
            if not matching_events.empty:
                item.properties["aorc:calendar_year_rank"] = int(matching_events.iloc[0]["annual_rank"])
                item.properties["aorc:collection_rank"] = int(matching_events.iloc[0]["por_rank"])

        collection = self.sort_collection(collection_id, "aorc:collection_rank")
        collection.add_asset(
            "valid-transposition-region",
            pystac.Asset(
                href=self.valid_transposition_region.self_href,
                title="Valid Transposition Region",
                description=f"Valid transposition region for {self.watershed.id} watershed",
                media_type=pystac.MediaType.GEOJSON,
                roles=["valid_transposition_region"],
            ),
        )

        collection.add_asset(
            "watershed",
            pystac.Asset(
                href=self.watershed.self_href,
                title="Watershed",
                description=f"{self.watershed.id} watershed",
                media_type=pystac.MediaType.GEOJSON,
                roles=["watershed"],
            ),
        )

        collection.save_object(dest_href=self.spm.collection_file(collection_id), include_self_link=False)
        self.add_collection_to_catalog(collection, override=True)
        self.sanitize_catalog_assets()
        return collection


def storm_search(
    catalog: StormCatalog,
    storm_start_date: datetime,
    storm_duration_hours: int,
    por_rank: int = None,
    return_item: bool = False,
    scale_max: float = 12.0,
    collection_id: str = None,
) -> Union[dict, AORCItem]:
    """
    Search for a storm event.

    Args:
        catalog (StormCatalog): The storm catalog.
        storm_start_date (datetime): The start date of the storm.
        storm_duration_hours (int): The duration of the storm in hours.
        return_item (bool): Whether to return the storm item.
        scale_max (float): The maximum scale for the thumbnail.
        collection_id (str): The ID of the collection.

    Returns
    -------
        Union[dict, AORCItem]: The storm search results or the storm item.
    """
    if not collection_id:
        collection_id = catalog.spm.storm_collection_id(storm_duration_hours)
    watershed = catalog.watershed
    valid_transposition_domain = catalog.valid_transposition_region

    logging.debug(
        "%s: searching %s - for max %d hr event.",
        storm_start_date.strftime("%Y-%m-%dT%H"),
        watershed.id,
        storm_duration_hours,
    )
    if por_rank:
        item_id = f"{por_rank}"
    else:
        item_id = f"{storm_start_date.strftime('%Y-%m-%dT%H')}"
    item_dir = catalog.spm.collection_item_dir(collection_id, item_id)

    event_item = AORCItem(
        item_id,
        storm_start_date,
        timedelta(hours=storm_duration_hours),
        shape(watershed.geometry),
        shape(valid_transposition_domain.geometry),
        item_dir,
        watershed.id,
        valid_transposition_domain.id,
        href=catalog.spm.collection_item(collection_id, item_id),
    )

    _, _, event_stats, centroid = event_item.max_transpose()
    logging.debug("Centroid: %s", centroid)
    logging.debug("Statistics: %s", event_stats)
    logging.debug("Storm Date: %s", storm_start_date.strftime("%Y-%m-%dT%H"))
    logging.debug("Destination href: %s", catalog.spm.collection_item(collection_id, event_item.id))
    if return_item:
        if not os.path.exists(item_dir):
            os.makedirs(item_dir)
        event_item.aorc_thumbnail(scale_max=scale_max)
        event_item.max_precip_point()
        event_item.save_object(dest_href=catalog.spm.collection_item(collection_id, event_item.id))
        return event_item
    else:
        return {
            "storm_date": storm_start_date.strftime("%Y-%m-%dT%H"),
            "centroid": centroid,
            "aorc:statistics": event_stats,
        }


def serial_processor(
    func: callable,
    catalog: StormCatalog,
    storm_duration: int,
    output_csv: str,
    event_dates: list[datetime],
    with_tb: bool = False,
):
    """
    Run function in serial using only one processor.

    Args:
        func (callable): The function to run.
        catalog (StormCatalog): The storm catalog.
        storm_duration (int): The duration of the storm.
        output_csv (str): Path to the output CSV file.
        event_dates (list[datetime]): List of event dates.
        with_tb (bool): Whether to include traceback in error logs.
    """
    if not os.path.exists(output_csv):
        with open(output_csv, "w", encoding="utf-8") as f:
            f.write("storm_date,min,mean,max,x,y\n")

    count = len(event_dates)

    with open(output_csv, "a", encoding="utf-8") as f:
        for date in event_dates:
            try:
                r = func(catalog, date, storm_duration)
                f.write(storm_search_results_to_csv_line(r))
                logging.info("%s processed (%d remaining)", r["storm_date"], count)
                count -= 1
            except Exception as e:
                if with_tb:
                    tb = traceback.format_exc()
                    logging.error("Error processing: %s\n%s", e, tb)
                else:
                    logging.error("Error processing: %s", e)


def multi_processor(
    func: callable,
    catalog: StormCatalog,
    storm_duration: int,
    output_csv: str,
    event_dates: list[datetime],
    num_workers: int = None,
    use_threads: bool = False,
    with_tb: bool = False,
):
    """
    Run function in parallel using multiple processors or threads.

    TODO: Consider using this for `storm_search` in creating items as well as collecting event stats.

    Args:
        func (callable): The function to run.
        catalog (StormCatalog): The storm catalog.
        storm_duration (int): The duration of the storm.
        output_csv (str): Path to the output CSV file.
        event_dates (list[datetime]): List of event dates.
        num_workers (int, optional): Number of workers to use.
        use_threads (bool): Whether to use threads instead of processes.
        with_tb (bool): Whether to include traceback in error logs.
    """
    if use_threads:
        executor = ThreadPoolExecutor
    else:
        executor = ProcessPoolExecutor

    if not os.path.exists(output_csv):
        # append_mode=True
        with open(output_csv, "w", encoding="utf-8") as f:
            f.write("storm_date,min,mean,max,x,y\n")

    count = len(event_dates)

    with open(output_csv, "a", encoding="utf-8") as f:
        with executor(max_workers=num_workers) as executor:
            futures = [executor.submit(func, catalog, date, storm_duration) for date in event_dates]
            for future in as_completed(futures):
                count -= 1
                try:
                    r = future.result()
                    f.write(storm_search_results_to_csv_line(r))
                    logging.info("%s processed (%d remaining)", r["storm_date"], count)

                except Exception as e:
                    if with_tb:
                        tb = traceback.format_exc()
                        logging.error("Error processing: %s\n%s", e, tb)
                        continue
                    else:
                        logging.error("Error processing: %s", e)
                        continue


def collect_event_stats(
    event_dates: list[datetime],
    catalog: StormCatalog,
    collection_id: str = None,
    storm_duration: int = 72,
    num_workers: int = None,
    use_threads: bool = False,
    with_tb: bool = False,
    use_parallel_processing: bool = True,
):
    """
    Collect statistics for storm events.

    Args:
        event_dates (list[datetime]): List of event dates.
        catalog (StormCatalog): The storm catalog.
        collection_id (str, optional): The ID of the collection.
        storm_duration (int): The duration of the storm.
        num_workers (int, optional): Number of workers to use.
        use_threads (bool): Whether to use threads instead of processes.
        with_tb (bool): Whether to include traceback in error logs.
        use_parallel_processing (bool): Whether to process storm stats using parallel processing.
    """
    if not collection_id:
        collection_id = catalog.spm.storm_collection_id(storm_duration)

    collection_dir = catalog.spm.collection_dir(collection_id)

    if not os.path.exists(collection_dir):
        os.makedirs(collection_dir)

    if not num_workers and not use_threads:
        if os.cpu_count() > 2:
            num_workers = os.cpu_count() - 2
        else:
            num_workers = 1
    elif not num_workers and use_threads:
        num_workers = 15

    output_csv = os.path.join(collection_dir, "storm-stats.csv")
    if use_parallel_processing:
        logging.info("Using %s cpu's for collecting event stats", num_workers)
        multi_processor(
            func=storm_search,
            catalog=catalog,
            storm_duration=storm_duration,
            output_csv=output_csv,
            event_dates=event_dates,
            num_workers=num_workers,
            use_threads=use_threads,
            with_tb=with_tb,
        )
    else:
        logging.info("Processing event stats serially.")
        serial_processor(
            func=storm_search,
            catalog=catalog,
            storm_duration=storm_duration,
            output_csv=output_csv,
            event_dates=event_dates,
            with_tb=with_tb,
        )


def create_items(
    event_dates: list[dict],
    catalog: StormCatalog,
    collection_id: str = None,
    storm_duration: int = 72,
    num_workers: int = None,
    with_tb: bool = False,
) -> List:
    """
    Create items for storm events, setting the item ID to `por_rank` instead of storm_date.

    Args:
        event_dates (list[dict]): List of event metadata (includes storm_date & por_rank).
        catalog (StormCatalog): The storm catalog.
        collection_id (str, optional): The ID of the collection.
        storm_duration (int): The duration of the storm.
        num_workers (int, optional): Number of workers to use.
        with_tb (bool): Whether to include traceback in error logs.

    Returns
    -------
        list: List of created event items.
    """
    event_items = []

    if not collection_id:
        collection_id = catalog.spm.storm_collection_id(storm_duration)

    collection_dir = catalog.spm.collection_dir(collection_id)
    if not os.path.exists(collection_dir):
        os.makedirs(collection_dir)

    count = len(event_dates)

    if not num_workers:
        num_workers = os.cpu_count()

    storm_data = [(e["storm_date"], e["por_rank"]) for e in event_dates]

    with ProcessPoolExecutor(max_workers=num_workers) as executor:
        futures = [
            executor.submit(
                storm_search,
                catalog,
                storm_date,
                storm_duration,
                por_rank=por_rank,
                collection_id=collection_id,
                return_item=True,
            )
            for storm_date, por_rank in storm_data
        ]

        for future in as_completed(futures):
            count -= 1
            try:
                r = future.result()
                logging.info("%s processed (%d remaining)", r.datetime, count)
                event_items.append(r)

            except Exception as e:
                if with_tb:
                    tb = traceback.format_exc()
                    logging.error("Error processing: %s\n%s", e, tb)
                else:
                    logging.error("Error processing: %s", e)

    return event_items


def init_storm_catalog(
    catalog_id: str, config: dict, local_catalog_dir: str, create_valid_transposition_region: bool = False
) -> pystac.Catalog:
    """
    Initialize a storm catalog.

    Args:
        catalog_id (str): The ID of the catalog.
        config (dict): Configuration dictionary.
        local_catalog_dir (str): Local directory for the catalog.
        create_valid_transposition_region (bool): Whether to create a valid transposition region.

    Returns
    -------
        pystac.Catalog: The initialized catalog.
    """
    watershed_config = config.get("watershed")
    tr_config = config.get("transposition_region")

    spm = StacPathManager(local_catalog_dir)

    if not os.path.exists(spm.catalog_dir):
        os.makedirs(spm.catalog_dir, exist_ok=True)

    logging.info("Creating `transposition_region` item for catalog: %s", catalog_id)
    transposition_region = HydroDomain(
        item_id=tr_config.get("id"),
        geometry=tr_config.get("geometry_file"),
        hydro_domain_type="transposition_region",
        description=tr_config.get("description"),
        title="Transposition Region",
    )
    transposition_region.save_object(dest_href=spm.catalog_asset(tr_config.get("id")), include_self_link=False)

    logging.info("Creating `watershed` item for catalog: %s", catalog_id)
    watershed = HydroDomain(
        item_id=watershed_config.get("id"),
        geometry=watershed_config.get("geometry_file"),
        hydro_domain_type="watershed",
        description=watershed_config.get("description"),
        title="Watershed",
    )
    watershed.save_object(dest_href=spm.catalog_asset(watershed_config.get("id")), include_self_link=False)

    if create_valid_transposition_region:
        logging.info("Creating `valid_transposition_region` item for catalog: %s", catalog_id)
        vtr_polygon = valid_spaces_item(watershed, transposition_region)
        vtr_id = f"{tr_config.get('id')}_valid"
        vtr = HydroDomain(
            item_id=vtr_id,
            geometry=vtr_polygon,
            hydro_domain_type="valid_transposition_region",
            description=f"Valid transposition region for {watershed.id} watershed",
            title="Valid Transposition Region",
        )
        vtr.save_object(dest_href=spm.catalog_asset(vtr_id), include_self_link=False)
        return watershed, transposition_region, vtr

    return watershed, transposition_region


def get_item_from_catalog_link(links: list, link_title: str, spm: StacPathManager) -> Item:
    """
    Get an item from the catalog links.

    Args:
        links (list): List of catalog links.
        link_title (str): The title of the link.
        spm (StacPathManager): The STAC path manager.

    Returns
    -------
        Item: The item from the catalog link.
    """
    matched_links = [link for link in links if link.title == link_title]
    if len(matched_links) == 0:
        return None
    if len(matched_links) > 1:
        raise ValueError(f"Multiple links found with title: {link_title}")

    relative_path = matched_links[0].target.replace("./", "")
    absolute_path = os.path.join(spm.catalog_dir, relative_path)

    absolute_path = os.path.abspath(absolute_path)

    item = pystac.read_file(absolute_path)
    if not isinstance(item, Item):
        raise ValueError(f"Expected an Item object at {absolute_path} not : {type(item)}")
    return item


def storm_search_results_to_csv_line(storm_search_results: dict) -> str:
    """
    Convert storm search results to a CSV line.

    Args:
        storm_search_results (dict): The storm search results.

    Returns
    -------
        str: The CSV line.
    """
    event_stats = storm_search_results["aorc:statistics"]
    storm_date = storm_search_results["storm_date"]
    centroid = storm_search_results["centroid"]
    stats = f"{event_stats['min']},{event_stats['mean']},{event_stats['max']}"
    return f"{storm_date},{stats},{centroid.x},{centroid.y}\n"


def find_missing_storm_dates(file_path: str, start_date: str, stop_date: str, every_n_hours: int) -> List:
    """
    Find missing storm dates in a CSV file.

    Args:
        file_path (str): Path to the CSV file.
        start_date (str): The start date.
        stop_date (str): The stop date.
        every_n_hours (int): The interval in hours.

    Returns
    -------
        list: List of missing storm dates.
    """
    df = pd.read_csv(file_path)
    df["storm_date"] = pd.to_datetime(df["storm_date"], format="%Y-%m-%dT%H")
    logging.info("Loaded %d storm events from %s", len(df), file_path)

    start_datetime = datetime.strptime(start_date, "%Y-%m-%d")
    stop_datetime = datetime.strptime(stop_date, "%Y-%m-%d")
    duration = timedelta(hours=every_n_hours)

    complete_range = pd.date_range(start=start_datetime, end=stop_datetime, freq=duration)
    logging.info("Expecting %d storm events for %s - %s", len(complete_range), start_date, stop_date)

    existing_datetimes = set(df["storm_date"])
    missing_datetimes = [dt for dt in complete_range if dt not in existing_datetimes]

    return missing_datetimes


def new_catalog(
    catalog_id: str,
    config_file: str,
    local_directory: str = None,
    catalog_description: str = "",
) -> StormCatalog:
    """
    Create a new storm catalog.

    Args:
        catalog_id (str): The ID of the catalog.
        config_file (str): Path to the configuration file.
        local_directory (str, optional): Local directory for the catalog.
        catalog_description (str): Description of the catalog.

    Returns
    -------
        StormCatalog: The new storm catalog.
    """
    # Step 1: Load config
    with open(config_file, "r", encoding="utf-8") as f:
        config = json.load(f)
    validate_config(config)

    # Step 2: Create catalog
    if not local_directory:
        local_directory = os.getcwd()

    # Utility class to force and manage paths
    spm = StacPathManager(os.path.join(local_directory, catalog_id))

    watershed, transposition_region, vtr = init_storm_catalog(
        catalog_id, config, spm.catalog_dir, create_valid_transposition_region=True
    )

    storm_catalog = StormCatalog(
        catalog_id,
        watershed,
        transposition_region,
        catalog_description,
        spm.catalog_dir,
        valid_transposition_region=vtr,
    )

    storm_catalog.save()
    logging.info("Catalog has been created.")
    return storm_catalog


def new_collection(
    catalog: Union[str | StormCatalog],
    start_date: str = "1979-02-01",
    end_date: str = None,
    storm_duration: int = 72,
    min_precip_threshold: int = 1,
    top_n_events: int = 5,
    check_every_n_hours: int = 6,
    specific_dates: list = None,
    num_workers: int = None,
    with_tb: bool = False,
    create_new_items: bool = True,
):
    """
    Create a new storm collection.

    Args:
        catalog (Union[str | StormCatalog]): The storm catalog or path to the catalog file.
        start_date (str): The start date for the collection.
        end_date (str, optional): The end date for the collection.
        storm_duration (int): The duration of the storm.
        min_precip_threshold (int): The minimum precipitation threshold.
        top_n_events (int): The number of top events to include.
        check_every_n_hours (int): The interval in hours to check for storms.
        specific_dates (list, optional): Specific dates to include.
        num_workers (int, optional): Number of cpu's to use during processing.
        with_tb (bool): Whether to include traceback in error logs.
        create_new_items (bool): Create items (or skip if items exist)
    """
    initialize_logger()

    if isinstance(catalog, str):
        storm_catalog = StormCatalog.from_file(catalog)
    elif isinstance(catalog, StormCatalog):
        storm_catalog = catalog
    else:
        raise ValueError(f"Catalog must be a path to a catalog file or a StormCatalog object not {type(catalog)}")

    if not end_date:
        end_date = datetime.now().strftime("%Y-%m-%dT%H")

    # logging.info(f"specific_dates: {specific_dates}")
    if not specific_dates:
        logging.info("Generating date range from %s to %s", start_date, end_date)
        dates = generate_date_range(start_date, end_date, every_n_hours=check_every_n_hours)
    elif len(specific_dates) > 0:
        logging.debug("Using specific dates: %s", specific_dates)
        dates = specific_dates
    elif len(specific_dates) == 0:
        logging.info("No specific dates provided.")
        dates = None
    else:
        logging.error("Unrecognized specific_dates argument or related  error.")
        dates = None

    collection_id = storm_catalog.spm.storm_collection_id(storm_duration)
    logging.info("Creating collection `%s` for period %s - %s", collection_id, start_date, end_date)

    if dates:
        logging.info("Collecting event stats for %d dates", len(dates))
        collect_event_stats(
            dates, storm_catalog, collection_id, storm_duration, num_workers=num_workers, with_tb=with_tb
        )
    stats_csv = os.path.join(storm_catalog.spm.collection_dir(collection_id), "storm-stats.csv")
    try:
        logging.info("Starting storm analysis for: %s", stats_csv)
        analyzer = StormAnalyzer(stats_csv, min_precip_threshold, storm_duration)
    except ValueError as e:
        logging.error("No events at threshold `min_precip_threshold` %d: %s", min_precip_threshold, e)
        return

    ranked_data = analyzer.rank_and_save(collection_id, storm_catalog.spm)

    top_events = ranked_data[ranked_data["por_rank"] <= top_n_events].copy()

    if create_new_items:
        event_items = create_items(
            top_events.to_dict(orient="records"), storm_catalog, storm_duration=storm_duration, with_tb=with_tb
        )
        collection = storm_catalog.new_collection_from_items(collection_id, event_items)

    else:
        collection = storm_catalog.add_rank_to_collection(collection_id, top_events)

    collection.add_summary_stats(storm_catalog.spm)
    collection.watershed_centroid_feature_collection(storm_catalog.spm)
    collection.max_precip_feature_collection(storm_catalog.spm)

    storm_catalog.add_collection_to_catalog(collection, override=True)
    storm_catalog.save_catalog()
    return collection


def resume_collection(
    catalog: str,
    start_date: str = "1979-02-01",
    end_date: str = None,
    storm_duration: int = 24,
    min_precip_threshold: int = 1,
    top_n_events: int = 5,
    check_every_n_hours: int = 6,
    num_workers: int = None,
    with_tb: bool = False,
    create_items: bool = True,
):
    """
    Resume a storm collection.

    Args:
        catalog (str): Path to the catalog file.
        start_date (str): The start date for the collection.
        end_date (str, optional): The end date for the collection.
        storm_duration (int): The duration of the storm.
        min_precip_threshold (int): The minimum precipitation threshold.
        top_n_events (int): The number of top events to include.
        check_every_n_hours (int): The interval in hours to check for storms.
        num_workers (int, optional): Number of cpu's to use during processing.
        with_tb (bool): Whether to include traceback in error logs.
    """
    initialize_logger()
    storm_catalog = StormCatalog.from_file(catalog)

    if not end_date:
        end_date = datetime.now().strftime("%Y-%m-%d")

    collection_id = storm_catalog.spm.storm_collection_id(storm_duration)
    partial_stats_csv = os.path.join(storm_catalog.spm.collection_dir(collection_id), "storm-stats.csv")
    logging.info("Searching for missing storm dates in %s", partial_stats_csv)
    dates = find_missing_storm_dates(partial_stats_csv, start_date, end_date, every_n_hours=check_every_n_hours)
    logging.info("%d dates found missing from %s - %s.", len(dates), start_date, end_date)

    new_collection(
        catalog=storm_catalog,
        start_date=start_date,
        end_date=end_date,
        storm_duration=storm_duration,
        min_precip_threshold=min_precip_threshold,
        top_n_events=top_n_events,
        check_every_n_hours=check_every_n_hours,
        specific_dates=dates,
        num_workers=num_workers,
        with_tb=with_tb,
        create_new_items=create_items,
    )



"""Plot functions for USGS gage data."""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.stats as stats
from stormhub.hydro.utils import log_pearson_iii


def plot_ams_seasonal(df, site: str, save_to: str):
    """Rank and Plot annual maxima series statistics, colored by season."""
    # Map months to seasons
    df["season"] = ((df.index.month % 12 + 3) // 3).map({1: "Winter", 2: "Spring", 3: "Summer", 4: "Fall"})

    # Rank flows in ascending order
    df = df.sort_values("peak_va").reset_index()
    df["rank"] = range(1, len(df) + 1)

    # Define custom colors for seasons
    season_colors = {
        "Winter": "blue",
        "Spring": "green",
        "Summer": "orange",
        "Fall": "brown",
    }

    # Create scatter plot
    fig, ax = plt.subplots()
    for season, color in season_colors.items():
        subset = df[df["season"] == season]
        ax.scatter(subset["rank"], subset["peak_va"], color=color, label=season, edgecolor="black", linewidth=0.5, s=50)

    ax.set_xlabel("Rank (Low to High Flow)")
    ax.set_ylabel("Flow (cfs)")
    ax.set_title(f"Gage ID: {site} | Flow Ranked from Low to High")
    ax.legend(title="Season")
    ax.grid(True, linestyle="--", alpha=0.6)

    fig.savefig(save_to, bbox_inches="tight")
    plt.close()


def plot_ams(df: pd.DataFrame, site: str, save_to: str):
    """Plot annual maxima series statistics."""
    fig, ax = plt.subplots()
    ax.scatter(df.index, df["peak_va"], color="blue", edgecolor="black", linewidth=0.5, s=25)

    ax.set_xlabel("Annual Maxima Series")
    ax.set_ylabel("Flow (cfs)")
    ax.set_title(f"Gage ID: {site} | Peak FLows")
    ax.grid(True, linestyle="--", alpha=0.6)
    fig.savefig(save_to, bbox_inches="tight")
    plt.close()


def plot_nwis_statistics(stats_df: pd.DataFrame, site: str, save_to: str):
    """Plot streamflow statistics from NWIS."""
    # Ensure data is sorted by day of the year
    stats_df = stats_df.sort_values(by=["month_nu", "day_nu"])

    # Create x-axis as day of year
    stats_df["day_of_year"] = stats_df["month_nu"] * 30 + stats_df["day_nu"]  # Approximate day of year

    fig, ax = plt.subplots(figsize=(12, 6))

    # Plot percentiles
    ax.fill_between(
        stats_df["day_of_year"],
        stats_df["p05_va"],
        stats_df["p95_va"],
        color="lightblue",
        alpha=0.4,
        label="5th-95th Percentile",
    )
    ax.fill_between(
        stats_df["day_of_year"],
        stats_df["p10_va"],
        stats_df["p90_va"],
        color="blue",
        alpha=0.4,
        label="10th-90th Percentile",
    )
    ax.fill_between(
        stats_df["day_of_year"],
        stats_df["p25_va"],
        stats_df["p75_va"],
        color="darkblue",
        alpha=0.4,
        label="25th-75th Percentile",
    )

    # Plot mean values
    ax.plot(
        stats_df["day_of_year"],
        stats_df["mean_va"],
        marker="o",
        markersize=2,
        linestyle="-",
        color="black",
        label="Mean Flow",
    )

    # Labels and title
    ax.set_xlabel("Day of Year")
    ax.set_ylabel("Flow (cfs)")
    ax.set_title(f"Flow Statistics for USGS Site {site}")
    ax.legend()
    ax.grid(True, linestyle="--", alpha=0.5)

    fig.savefig(save_to, bbox_inches="tight")
    plt.close()


def plot_log_pearson_iii(
    peak_flows: pd.Series,
    gage_id: str,
    save_to: str,
):
    """
    Plot the return period (recurrence interval) vs. peak flow using Log-Pearson Type III analysis.

    Parameters
    ----------
        peak_flows (pd.Series): List or array of peak flow values.
    """
    standard_return_periods = [2, 5, 10, 25, 50, 100, 500]
    lp3_estimates = log_pearson_iii(peak_flows, standard_return_periods)
    lp3_values = lp3_estimates.values()

    fig, ax = plt.subplots(figsize=(10, 6))
    ax.plot(
        standard_return_periods,
        lp3_values,
        marker="o",
        linestyle="-",
        color="blue",
        label="Log-Pearson III Fit",
        zorder=2,
    )

    # Set predefined tick marks
    ax.set_xscale("log")  # Log scale for return period
    ax.set_xticks(standard_return_periods)  # Set ticks at predefined return periods
    ax.set_xticklabels([str(rp) for rp in standard_return_periods])  # Force display of exact labels

    # Labels and title
    ax.set_xlabel("Return Period (years)")
    ax.set_ylabel("Peak Flow (cfs)")
    ax.set_title(f"{gage_id} | Log-Pearson Type III Estimates \n(No regional skew))")
    ax.grid(True, which="both", linestyle="--", alpha=0.6)
    ax.legend()

    fig.savefig(save_to, bbox_inches="tight")
    plt.close()



"""Creates grid file from metadata json created by extract_dss_metadata.py and add_coords_to_metadata.py in format in the following way.

{
  "19790222_72hr_st1_r265.dss": {
    "precip_pathname": "/SHG4K/TRINITY/PRECIPITATION/22FEB1979:0000/22FEB1979:0100/AORC/",
    "temperature_pathname": "/SHG4K/TRINITY/TEMPERATURE/22FEB1979:0100//AORC/",
    "storm_type": "st1",
    "rank": "r265",
    "storm_center_lon": -90.9016,
    "storm_center_lat": 31.1662
  },
  "19790302_72hr_st1_r118.dss": {
    "precip_pathname": "/SHG4K/TRINITY/PRECIPITATION/02MAR1979:0000/02MAR1979:0100/AORC/",
    "temperature_pathname": "/SHG4K/TRINITY/TEMPERATURE/02MAR1979:0100//AORC/",
    "storm_type": "st1",
    "rank": "r118",
    "storm_center_lon": -87.535,
    "storm_center_lat": 32.4662
  }.
"""

import json
from pyproj import Transformer
from datetime import datetime
import logging
from stormhub.logger import initialize_logger

initialize_logger()


def parse_date_from_path(path: str):
    """Parse date from dss path."""
    try:
        return datetime.strptime(path.strip("/").split("/")[3], "%d%b%Y:%H%M").date()
    except Exception:
        return None


def transform_coords(lon, lat, transformer):
    """Transform given coords and return formatted storm center line."""
    x, y = transformer.transform(lon, lat)
    return f"     Storm Center X: {x}\n", f"     Storm Center Y: {y}\n"


def build_lines(dss_data, transformer, defaults):
    """Build grid file lines."""
    lines = [
        "Grid Manager: T Transpose\n",
        "     Version: 4.11\n",
        "     Filepath Separator: \\\n",
        "End:\n\n",
    ]

    for dss_file, meta in sorted(
        dss_data.items(), key=lambda item: int(item[1].get("rank", "r0").lstrip("r"))
    ):  # sort by rank
        precip_path = meta.get("precip_pathname")
        temp_path = meta.get("temperature_pathname")
        storm_date = parse_date_from_path(precip_path or temp_path)
        dss_filename = f"data\\{dss_file}"

        if not storm_date:
            logging.error(f"Could not parse date for {dss_file}")
            continue

        for grid_type, path in [("Precipitation", precip_path), ("Temperature", temp_path)]:
            lines.append(f"Grid: {dss_file.replace('.dss', '')}\n")
            lines.append(f"     Grid Type: {grid_type}\n")
            lines.append(f"     Last Modified Date: {defaults['modified_date']}\n")
            lines.append(f"     Last Modified Time: {defaults['modified_time']}\n")
            lines.append(f"     Reference Height Units: {defaults['ref_units']}\n")
            lines.append(f"     Reference Height: {defaults['ref_height']}\n")
            lines.append("     Data Source Type: External DSS\n")
            lines.append(f"     Variant: {defaults['variant']}\n")
            lines.append(f"       Last Variant Modified Date: {defaults['modified_date']}\n")
            lines.append(f"       Last Variant Modified Time: {defaults['modified_time']}\n")
            lines.append("       Default Variant: Yes\n")
            lines.append(f"       DSS File Name: {dss_filename}\n")
            lines.append(f"       DSS Pathname: {path}\n")
            lines.append(f"     End Variant: {defaults['variant']}\n")
            lines.append(f"     Use Lookup Table: {defaults['use_lookup']}\n")

            if "storm_center_lon" in meta and "storm_center_lat" in meta:
                x_line, y_line = transform_coords(meta["storm_center_lon"], meta["storm_center_lat"], transformer)
                lines.append(x_line)
                lines.append(y_line)

            lines.append("End:\n\n")

    return lines


if __name__ == "__main__":
    metadata_path = "dss_metadata.json"
    output_grid_path = "T_Transpose_recreated.grid"

    output_crs = (
        'PROJCRS["USA_Contiguous_Albers_Equal_Area_Conic_USGS_version",'
        'BASEGEOGCRS["NAD83",DATUM["North American Datum 1983",'
        'ELLIPSOID["GRS 1980",6378137,298.257222101,LENGTHUNIT["metre",1]],'
        'ID["EPSG",6269]],PRIMEM["Greenwich",0,ANGLEUNIT["Degree",0.0174532925199433]]],'
        'CONVERSION["unnamed",METHOD["Albers Equal Area",ID["EPSG",9822]],'
        'PARAMETER["Latitude of false origin",23,ANGLEUNIT["Degree",0.0174532925199433],ID["EPSG",8821]],'
        'PARAMETER["Longitude of false origin",-96,ANGLEUNIT["Degree",0.0174532925199433],ID["EPSG",8822]],'
        'PARAMETER["Latitude of 1st standard parallel",29.5,ANGLEUNIT["Degree",0.0174532925199433],ID["EPSG",8823]],'
        'PARAMETER["Latitude of 2nd standard parallel",45.5,ANGLEUNIT["Degree",0.0174532925199433],ID["EPSG",8824]],'
        'PARAMETER["Easting at false origin",0,LENGTHUNIT["US survey foot",0.304800609601219],ID["EPSG",8826]],'
        'PARAMETER["Northing at false origin",0,LENGTHUNIT["US survey foot",0.304800609601219],ID["EPSG",8827]]],'
        'CS[Cartesian,2],AXIS["(E)",east,ORDER[1],LENGTHUNIT["US survey foot",0.304800609601219,ID["EPSG",9003]]],'
        'AXIS["(N)",north,ORDER[2],LENGTHUNIT["US survey foot",0.304800609601219,ID["EPSG",9003]]]]'
    )

    defaults = {
        "modified_date": "23 May 2025",
        "modified_time": "16:30:15",
        "ref_units": "Meters",
        "ref_height": 10.0,
        "variant": "Variant-1",
        "use_lookup": "No",
    }

    transformer = Transformer.from_crs("EPSG:4326", output_crs, always_xy=True)

    with open(metadata_path) as f:
        dss_data = json.load(f)

    lines = build_lines(dss_data, transformer, defaults)

    with open(output_grid_path, "w") as f:
        f.writelines(lines)

    logging.info(f"Created: {output_grid_path}")



"""Hydrodomain Item class."""

from datetime import datetime
from typing import Any, Union
import logging
import fiona.errors
import geopandas as gpd
from pystac import Item
from pystac.extensions.projection import ProjectionExtension
from shapely.geometry import Polygon, mapping, shape, MultiPolygon

HYDRO_DOMAIN_DESCRIPTION = "hydro_domain:description"
HYDRO_DOMAIN_TYPE = "hydro_domain:type"
DATETIME_INFO = "datetime_info"
PROJ_EPSG = "proj:epsg"


class HydroDomain(Item):
    """
    Initialize a hydrological domain Item.

    Args:
        item_id (str): The ID of the Item.
        geometry (str | Polygon): The Item geometry.
        hydro_domain_type (str): Hydrological domain type. Options include 'watershed', 'transposition_region', and 'valid_transposition_region'.
        relevant_datetime (str | datetime, optional): Datetime used for the item. If one is not provided then the item creation time is used.
        relevant_datetime_description (str): Description of the datetime.
        **kwargs (Any): Additional keyword arguments.
    """

    def __init__(
        self,
        item_id: str,
        geometry: str | Polygon,
        hydro_domain_type: str,
        relevant_datetime: str | datetime = None,
        relevant_datetime_description: str = None,
        **kwargs: Any,
    ):
        self.item_id = item_id
        self.geometry = self.load_geometry(geometry)
        if hydro_domain_type not in ["watershed", "transposition_region", "valid_transposition_region"]:
            raise ValueError(
                f"hydro_domain_type must be one of: watershed, transposition_region, valid_transposition_region, not {hydro_domain_type}"
            )

        self.hydro_domain_type = hydro_domain_type

        if isinstance(relevant_datetime, str):
            self.relevant_datetime = datetime.fromisoformat(relevant_datetime)
        elif isinstance(relevant_datetime, datetime):
            self.relevant_datetime = relevant_datetime
        else:
            self.relevant_datetime = datetime.now()

        if not relevant_datetime_description:
            self.relevant_datetime_description = "Creation date of the item"

        self.stac_extensions = kwargs.get("stac_extensions", None)
        self.href = kwargs.get("href", None)
        self.properties = kwargs.get("properties", {})
        self.start_datetime = kwargs.get("start_datetime", None)
        self.end_datetime = kwargs.get("end_datetime", None)
        self.collection = kwargs.get("collection", None)
        self.extra_fields = kwargs.get("extra_fields", None)
        self.assets = kwargs.get("assets", None)
        self.description = kwargs.get("description", None)
        self.title = kwargs.get("title", None)

        if self.title:
            self.properties["title"] = self.title
        if self.description:
            self.properties["description"] = self.description

        super().__init__(
            id=self.item_id,
            geometry=mapping(self.geometry),
            bbox=self.geometry.bounds,
            datetime=self.relevant_datetime,
            properties=self.properties,
            stac_extensions=self.stac_extensions,
            href=self.href,
            collection=self.collection,
            extra_fields=self.extra_fields,
            assets=self.assets,
        )

        self._register_extensions()
        self.properties[HYDRO_DOMAIN_TYPE] = self.hydro_domain_type
        self.properties[DATETIME_INFO] = self.relevant_datetime_description
        self.properties[PROJ_EPSG] = 4326

    @classmethod
    def from_item(cls, item: Item) -> "HydroDomain":
        """Create a HydroDomain instance from a STAC item."""
        return cls(
            item_id=item.id,
            geometry=shape(item.geometry),
            hydro_domain_type=item.properties.get(HYDRO_DOMAIN_TYPE),
            relevant_datetime=item.properties.get("datetime"),
            relevant_datetime_description=item.properties.get("datetime_description"),
        )

    def _register_extensions(self) -> None:
        ProjectionExtension.add_to(self)
        # StorageExtension.add_to(self)

    def _ensure_datetime(self, dt: str) -> datetime:
        if isinstance(dt, str):
            return datetime.datetime.fromisoformat(dt)
        return dt or datetime.datetime.now()

    def load_geometry(self, geometry_source: Union[str, Polygon]) -> Polygon:
        """Load geometry from str or Polygon object."""
        if isinstance(geometry_source, str):
            try:
                gdf = gpd.read_file(geometry_source)
            except (fiona.errors.FionaValueError, fiona.errors.DriverError) as e:
                raise ValueError(f"Error reading the geometry file: {e}")
        elif isinstance(geometry_source, Polygon):
            gdf = gpd.GeoDataFrame(geometry=[geometry_source], crs="EPSG:4326")
        else:
            raise ValueError("geometry_source must be a file path or a Polygon object")

        if len(gdf) != 1:
            raise ValueError("The geometry must contain a single polygon")

        geometry = gdf.geometry.iloc[0]
        if isinstance(geometry, MultiPolygon) and len(geometry.geoms) == 1:
            logging.warning("Multipolygon type detected, attempting conversion to Polygon.")
            geometry = geometry.geoms[0]
            gdf.at[gdf.index[0], "geometry"] = geometry

        if not isinstance(geometry, Polygon):
            raise ValueError("The geometry must contain a single polygon")

        try:
            if gdf.crs != "EPSG:4326":
                gdf = gdf.to_crs("EPSG:4326")
        except ValueError as e:
            raise ValueError(f"Error converting CRS to EPSG:4326: {e}")

        return gdf.geometry.iloc[0]



"""Creates dss files from a given watershed for each event in a metadata json formatted in the following way.

{
  "1988-09-13": {
    "event_id": "236",
    "st_number": "st5"
  },
  "1993-11-14": {
    "event_id": "190",
    "st_number": "st5"
  }
}.
"""

import os
import json
from datetime import datetime
from stormhub.met.zarr_to_dss import noaa_zarr_to_dss, NOAADataVariable


def load_storm_events(json_path: str) -> dict:
    """Load storm event metadata from a JSON file."""
    with open(json_path, "r") as f:
        return json.load(f)


def process_storm_events(storms: dict, dss_dir: str, aoi_path: str, aoi_name: str) -> None:
    """Process each storm event and generate DSS files."""
    for date, attrs in storms.items():
        storm_start_datetime = datetime.strptime(date, "%Y-%m-%d")
        storm_start_str = storm_start_datetime.strftime("%Y%m%d")
        output_name = f"{storm_start_str}_72h_{attrs['st_number']}_r{attrs['event_id']}.dss"
        output_dss_path = os.path.join(dss_dir, output_name)

        if os.path.exists(output_dss_path):
            print(f"Skipping creation for {output_dss_path} as it already exists.")
            continue

        noaa_zarr_to_dss(
            output_dss_path=output_dss_path,
            aoi_geometry_gpkg_path=aoi_path,
            aoi_name=aoi_name,
            storm_start=storm_start_datetime,
            variable_duration_map={
                NOAADataVariable.TMP: 864,
                NOAADataVariable.APCP: 72,
            },
        )


if __name__ == "__main__":
    json_path = "storm_event.json"
    dss_dir = "dss_outputs"
    aoi_path = "trinity-transpo-area-v01.geojson"
    aoi_name = "TRINITY"

    storms = load_storm_events(json_path)
    os.makedirs(dss_dir, exist_ok=True)
    process_storm_events(storms, dss_dir, aoi_path, aoi_name)



"""Create dss from aorc zarr data."""

from datetime import datetime, timedelta
from enum import Enum
import math
from typing import List, Tuple, Literal, Dict
from affine import Affine
from hecdss import HecDss, gridded_data
import numpy as np
from pandas import Timestamp
import geopandas as gpd
from geopandas import GeoDataFrame
import s3fs
import xarray as xr
from stormhub.met.consts import NOAA_AORC_S3_BASE_URL, KM_TO_M_CONVERSION_FACTOR, SHG_WKT
import logging


logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(message)s",
)


class MeasurementType(Enum):
    """Type of measurements."""

    PERCUM = "per_cum"
    INSTVAL = "inst_val"


class DSSPath:
    """Class for defining DSS paths."""

    def __init__(self, a_part: str, b_part: str, c_part: str, d_part: str, e_part: str, f_part: str):
        """
        Parameter descriptions from: https://www.hec.usace.army.mil/confluence/dssdocs/dssvueum/introduction/general-concepts-for-hec-dss.

        Args:
            a_part (str): Refers to the grid reference system. At present, GageInterp supports only the HRAP and SHG grid systems
            b_part (str): Contains the name of the region covered by the grid
            c_part (str): Refers to the parameter represented by the grid
            d_part (str): Contains the start time
            e_part (str): Contains the end time
            f_part (str): Refers to the version of the data
        """
        self.a_part = a_part
        self.b_part = b_part
        self.c_part = c_part
        self.d_part = d_part
        self.e_part = e_part
        self.f_part = f_part

    def __str__(self) -> str:
        """Return the DSS path string."""
        return f"/{self.a_part}/{self.b_part}/{self.c_part}/{self.d_part}/{self.e_part}/{self.f_part}/"


class NOAADataVariable(Enum):
    """Class of potential NOAA data variables to extract zarr data for."""

    APCP = "APCP_surface"
    DLWRF = "DLWRF_surface"
    DSWRF = "DSWRF_surface"
    PRES = "PRES_surface"
    SPFH = "SPFH_2maboveground"
    TMP = "TMP_2maboveground"
    UGRD = "UGRD_10maboveground"
    VGRD = "VGRD_10maboveground"

    @property
    def dss_variable_title(self) -> str:
        """Return variable title."""
        if self == NOAADataVariable.APCP:
            return "PRECIPITATION"
        elif self == NOAADataVariable.TMP:
            return "TEMPERATURE"
        else:
            return self.value

    @property
    def measurement_type(self) -> MeasurementType:
        """Return measurement type."""
        if self == NOAADataVariable.APCP:
            return MeasurementType.PERCUM
        else:
            return MeasurementType.INSTVAL

    @property
    def measurement_unit(self) -> str:
        """Return measurement units."""
        if self == NOAADataVariable.APCP:
            return "MM"
        elif self == NOAADataVariable.TMP:
            return "DEG C"
        else:
            raise NotImplementedError(f"Unit unknown for data variable {self.__repr__}")


def get_aorc_paths(storm_start: datetime, storm_end: datetime) -> list[str]:
    """Construct s3 paths for AORC dataset given storm start and end time."""
    aorc_paths = []
    if storm_start.year == storm_end.year:
        aorc_paths.append(f"{NOAA_AORC_S3_BASE_URL}/{storm_start.year}.zarr")
    else:
        for year in range(storm_start.year, storm_end.year + 1):  # plus one since end range is exclusive
            aorc_paths.append(f"{NOAA_AORC_S3_BASE_URL}/{year}.zarr")
    return aorc_paths


def date_range_dss_path_format(date: datetime, measurement_type: MeasurementType) -> Tuple[str, str]:
    """Format start and end times for DSS path based on a date and measurement type."""
    if measurement_type == MeasurementType.PERCUM:
        end_dt = date
        start_dt = end_dt - timedelta(hours=1)
        start_dt_str = start_dt.strftime("%d%b%Y:%H%M").upper()

        if end_dt.hour == 0 and end_dt.minute == 0:
            end_dt_str = start_dt.strftime("%d%b%Y:2400").upper()
        else:
            end_dt_str = end_dt.strftime("%d%b%Y:%H%M").upper()
    elif measurement_type == MeasurementType.INSTVAL:
        start_dt = date
        if start_dt.hour == 0 and start_dt.minute == 0:
            start_dt -= timedelta(days=1)
            start_dt_str = start_dt.strftime("%d%b%Y:2400").upper()
        else:
            start_dt_str = start_dt.strftime("%d%b%Y:%H%M").upper()
        end_dt_str = ""
    else:
        raise NotImplementedError(
            f"Start and end time definition from a single datetime is not defined for {measurement_type}."
        )

    return start_dt_str, end_dt_str


def get_lower_left_xy(
    data: xr.Dataset, resolution: int, x_coord_of_grid_cell_zero=0, y_coord_of_grid_cell_zero=0
) -> Tuple[int, int]:
    """Get lower left xy from gridded data."""
    y_coords = data.y.to_numpy()
    x_coords = data.x.to_numpy()

    y_coord = y_coords[-1] if y_coords[-1] < y_coords[0] else y_coords[0]
    x_coord = x_coords[-1] if x_coords[-1] < x_coords[0] else x_coords[0]

    affine_transform = Affine(resolution, 0.0, x_coord, 0.0, resolution, y_coord)

    cellsize = affine_transform[0]

    xmin = affine_transform[2] - x_coord_of_grid_cell_zero
    ymax = affine_transform[5] - y_coord_of_grid_cell_zero

    lower_left_x = int(math.floor(xmin / cellsize))
    lower_left_y = int(math.floor(ymax / cellsize))

    return (lower_left_x, lower_left_y)


def convert_temperature_dataset(data: xr.Dataset, chunk_size: int = 144) -> xr.Dataset:
    """Convert temperature in Kelvin to the desired output_unit. Utilizes chunking to save memory."""
    output_unit = NOAADataVariable.TMP.measurement_unit
    data_unit = data.units
    if data_unit != "K":
        raise ValueError(f"Expected temperature data in Kelvin, got measurement unit of {data_unit} instead")

    if output_unit != "K":
        data_shape = data.shape
        c_degrees_difference = np.full(data_shape, 273.15)
        num_chunks = (data_shape[0] + chunk_size - 1) // chunk_size

        converted_chunks = []

        for i in range(num_chunks):
            start = i * chunk_size
            end = min((i + 1) * chunk_size, data_shape[0])

            data_chunk = data.isel(time=slice(start, end))

            if output_unit == "DEG C":
                converted_chunk = data_chunk - c_degrees_difference[start:end]
            elif output_unit == "DEG F":
                c_data_chunk = data_chunk - c_degrees_difference[start:end]
                scale_difference = np.full(c_data_chunk.shape, 9 / 5)
                scale_data_chunk = c_data_chunk * scale_difference
                f_difference = np.full(c_data_chunk.shape, 32)
                converted_chunk = scale_data_chunk + f_difference
            else:
                raise ValueError(
                    f"Temperature conversion only supported from Kelvin (K) to Celsius (DEG C) or Fahrenheit (DEG F); got output unit of {output_unit} instead"
                )

            converted_chunks.append(converted_chunk)

        # Concatenate all converted chunks along the 'time' dimension
        data = xr.concat(converted_chunks, dim="time")

    return data


def create_gridded_data(
    path: DSSPath,
    data: np.ndarray,
    grid_type: Literal[
        "undefined_grid_with_time",
        "undefined_grid",
        "hrap_grid_with_time_ref",
        "hrap_grid",
        "albers_with_time_ref",
        "albers",
        "specified_grid_with_time_ref",
        "specified_grid",
    ],
    data_type: Literal["per_aver", "per_cum", "inst_val", "inst_cum", "freq", "invalid"],
    cell_size: float,
    data_units: str,
    srs_definition: str,
    lower_left_cell_x: int,
    lower_left_cell_y: int,
    x_coord_of_grid_cell_zero: float = 0,
    y_coord_of_grid_cell_zero: float = 0,
) -> gridded_data.GriddedData:
    """
    Create a gridded_data.GriddedData object from the HecDss library. Specifies default values for some parameters of GriddedData.

    Args:
        path: Path following the /part_a/part_b/part_c/part_d/part_e/part_f/ format
        data: A numpy array containing the gridded data values
        grid_type: The type of grid to use to display this data
        data_type: The measurement type of the data stored in the grid
        cell_size: The size of each grid cell
        data_units: The unit of measurement for the data
        srs_definition: The spatial reference system definition (either WKT or EPSG code)
        lower_left_cell_x: The x-coordinate index of the lower-left cell in the grid
        lower_left_cell_y: The y-coordinate index of the lower-left cell in the grid
    """
    number_of_ranges: int = 0
    time_zone_raw_offset: int = 0
    is_interval: int = 0
    is_time_stamped: int = 0
    data_source: str = ""
    srs_definition_type: int = 0
    srs_name: str = "WKT"
    time_zone_id: str = ""
    null_value: float = 0

    grid_type_map = {
        "undefined_grid_with_time": 400,
        "undefined_grid": 401,
        "hrap_grid_with_time_ref": 410,
        "hrap_grid": 411,
        "albers_with_time_ref": 420,
        "albers": 421,
        "specified_grid_with_time_ref": 430,
        "specified_grid": 431,
    }
    data_type_map = {"per_aver": 0, "per_cum": 1, "inst_val": 2, "inst_cum": 3, "freq": 4, "invalid": 5}

    gd = gridded_data.GriddedData.create(
        path=str(path),
        type=grid_type_map[grid_type],
        dataType=data_type_map[data_type],
        lowerLeftCellX=lower_left_cell_x,
        lowerLeftCellY=lower_left_cell_y,
        numberOfRanges=number_of_ranges,
        srsDefinitionType=srs_definition_type,
        timeZoneRawOffset=time_zone_raw_offset,
        isInterval=is_interval,
        isTimeStamped=is_time_stamped,
        dataUnits=data_units,
        dataSource=data_source,
        srsName=srs_name,
        srsDefinition=srs_definition,
        timeZoneID=time_zone_id,
        cellSize=cell_size,
        xCoordOfGridCellZero=x_coord_of_grid_cell_zero,
        yCoordOfGridCellZero=y_coord_of_grid_cell_zero,
        nullValue=null_value,
        data=data,
    )

    return gd


def interpolate_nan_values(ds: xr.DataArray) -> xr.DataArray:
    """Interpolate missing NaN values in a DataArray along the 'latitude' and 'longitude' dimensions using linear interpolation. Averages the results from both directions and fills remaining NaNs."""
    ds2_rechunked = ds.chunk({"latitude": -1, "longitude": -1})

    interpolated_lon = ds2_rechunked.interpolate_na(dim="longitude", method="linear")
    interpolated_lat = ds2_rechunked.interpolate_na(dim="latitude", method="linear")

    both_valid = (~interpolated_lon.isnull()) & (~interpolated_lat.isnull())
    average = (interpolated_lon + interpolated_lat) / 2

    interpolated_combined = xr.where(both_valid, average, interpolated_lon.combine_first(interpolated_lat))
    interpolated_combined.rio.write_crs(ds.rio.crs, inplace=True)

    interpolated_combined.attrs["units"] = "K"

    return interpolated_combined


def get_s3_zarr_data(
    s3_paths: List[str],
    aoi_gdf: GeoDataFrame,
    start_dt: datetime,
    end_dt: datetime,
    variables_of_interest: List[str],
    interp_nan_vals: bool = True,
) -> xr.Dataset:
    """Load and processe Zarr datasets from S3, clipping them to a given area of interest (AOI), filter by time and variables, and optionally interpolates missing values."""
    s3 = s3fs.S3FileSystem(anon=True)
    fileset = [s3fs.S3Map(root=path, s3=s3, check=False) for path in s3_paths]
    ds = xr.open_mfdataset(fileset, engine="zarr", chunks="auto", consolidated=True)

    # Select only variables of interest
    if variables_of_interest:
        ds = ds[variables_of_interest]

    # Reproject AOI and clip spatially
    aoi_gdf = aoi_gdf.to_crs(ds.rio.crs)
    aoi_shape = aoi_gdf.geometry.iloc[0]
    bounds = aoi_shape.bounds
    ds = ds.sel(
        time=slice(start_dt, end_dt), longitude=slice(bounds[0], bounds[2]), latitude=slice(bounds[1], bounds[3])
    )
    if interp_nan_vals:
        # Interpolate missing values for each variable
        for var in ds.data_vars:
            data_var = ds[var]

            # Compute valid mask
            valid_mask = ~data_var.isnull().all("time")

            # Determine which time slices need interpolation
            nan_mask = data_var.isnull() & valid_mask
            needs_interp = nan_mask.any(dim=["latitude", "longitude"]).compute()

            if not needs_interp.any():
                logging.info(f"All data for {var} is valid")
                continue

            # Interpolate time slices
            interpolated_slices = []
            for i, t in enumerate(data_var.time.values):
                if needs_interp[i]:
                    logging.info(f"Missing data for var {var} at time {t}. Interpolating...")
                    slice_ = data_var.sel(time=t)
                    interpolated = interpolate_nan_values(slice_)
                    interpolated_slices.append(interpolated.expand_dims(time=[t]))

            # Combine interpolated slices with original
            if interpolated_slices:
                interpolated_ds = xr.concat(interpolated_slices, dim="time")
                data_var = data_var.combine_first(interpolated_ds)

            ds[var] = data_var

    # Final spatial clip
    ds = ds.rio.clip([aoi_shape], drop=True, all_touched=True)

    return ds


def write_to_dss(
    output_dss_path: str,
    data: xr.Dataset,
    aoi_name: str,
    param_name: str,
    param_measurement_type: MeasurementType,
    param_measurement_unit: str,
    output_resolution_km: int,
    data_version: str,
):
    """
    Write geospatial data to a DSS file while transforming the data to fit DSS conventions.

    Args:
        output_dss_path: Path to the output DSS file
        zarr_data: An xarray dataset containing the geospatial data to be written to the DSS file
        aoi_name: The name of the area of interest (AOI)
        parameter_name: The name of the parameter being stored in the DSS file (e.g. "precipitation")
        parameter_measurement_type: The type of measurement type of the parameter
        output_resolution_km: The resolution for the data in km
        data_version: Represents where the data comes from (e.g. "AORC")
    """
    dss = HecDss(output_dss_path)
    output_resolution_m = output_resolution_km * KM_TO_M_CONVERSION_FACTOR

    logging.info(f"reprojecting dataset")
    times = data.time.values

    if len(times) <= 144:
        data: xr.DataArray = data.rio.reproject(SHG_WKT, resolution=output_resolution_m)
    else:
        # For larger datasets, chunking is used to avoid memory issues
        logging.info(f"Chunking dataset for reprojection")
        time_chunk_size = 144
        reprojected_chunks = []

        for i in range(0, len(times), time_chunk_size):
            chunk_times = times[i : i + time_chunk_size]
            chunk = data.sel(time=chunk_times)
            chunk = chunk.rio.reproject(SHG_WKT, resolution=output_resolution_m)
            reprojected_chunks.append(chunk)

        data = xr.concat(reprojected_chunks, dim="time")

    lower_x, lower_y = get_lower_left_xy(data, output_resolution_m)

    for time_step in data.time:
        time_step_data = data.sel(time=time_step)
        time_step_data = np.flipud(time_step_data.to_numpy())

        date = Timestamp(time_step.values).to_pydatetime()
        logging.info(f"processing timestep: {date}")
        start_dt_str, end_dt_str = date_range_dss_path_format(date, param_measurement_type)

        path = DSSPath(
            f"SHG{output_resolution_km}K",
            aoi_name.upper(),
            param_name.upper(),
            start_dt_str,
            end_dt_str,
            data_version.upper(),
        )

        gd = create_gridded_data(
            path=path,
            data=time_step_data,
            grid_type="albers_with_time_ref",
            data_type=param_measurement_type.value,
            cell_size=output_resolution_m,
            data_units=param_measurement_unit,
            srs_definition=SHG_WKT,
            lower_left_cell_x=lower_x,
            lower_left_cell_y=lower_y,
        )

        dss.put(gd)

    dss.close()


def noaa_zarr_to_dss(
    output_dss_path: str,
    aoi_geometry_gpkg_path: str,
    aoi_name: str,
    storm_start: datetime,
    variable_duration_map: Dict[NOAADataVariable, int],
):
    """Given a geometry and datetime information about a storm, writes variables of interest from NOAA dataset to DSS."""
    # arrange parameters
    all_variables = list(variable_duration_map.keys())
    min_start = storm_start + timedelta(hours=1)  # make exclusive
    max_end = storm_start + timedelta(hours=max(variable_duration_map.values()))
    aorc_paths = get_aorc_paths(min_start, max_end)
    aoi_gdf = gpd.read_file(aoi_geometry_gpkg_path)
    voi_keys = [v.value for v in all_variables]

    # get aorc data
    logging.info("Getting aorc data")
    aorc_data = get_s3_zarr_data(aorc_paths, aoi_gdf, min_start, max_end, voi_keys)
    logging.info("Successfully retrieved aorc data")

    # write to dss
    for data_variable, duration in variable_duration_map.items():
        var_start = storm_start + timedelta(hours=1)
        var_end = storm_start + timedelta(hours=duration)
        data = aorc_data[data_variable.value].sel(time=slice(var_start, var_end))

        if data_variable == NOAADataVariable.TMP:
            logging.info("converting temperature dataset")
            data = convert_temperature_dataset(data)
            logging.info("Successfully converted temperature dataset")
        logging.info("writing to dss")
        write_to_dss(
            output_dss_path=output_dss_path,
            data=data,
            aoi_name=aoi_name,
            param_name=data_variable.dss_variable_title,
            param_measurement_type=data_variable.measurement_type,
            param_measurement_unit=data_variable.measurement_unit,
            output_resolution_km=4,
            data_version="AORC",
        )



"""AORC Item class."""

import datetime
import json
import logging
import os
from typing import Any

import numpy as np
import s3fs
import xarray as xr
from affine import Affine
from matplotlib import patches
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from pyproj import CRS
from pystac import Asset, Item, MediaType
from pystac.extensions.projection import ProjectionExtension
from pystac.extensions.storage import CloudPlatform, StorageExtension
from shapely import Polygon, to_geojson
from shapely.geometry import shape

from stormhub.met.transpose import Transpose

NULL_POLYGON = Polygon()

from stormhub.met.consts import (
    AORC_PRECIP_VARIABLE,
    AORC_X_VAR,
    AORC_Y_VAR,
    MM_TO_INCH_CONVERSION_FACTOR,
    NOAA_AORC_S3_BASE_URL,
)


class AORCItem(Item):
    """Initialize an AORC Item.

    Args:
        item_id (str): The ID of the item.
        start_datetime (datetime.datetime): The AORC start datetime.
        duration_hours (int): Duration of AORC data in hours.
        watershed (str): Location of watershed geometry geojson.
        transposition_domain (str): Location of transposition domain geometry geojson.
        local_directory (str): Local directory path
        watershed_name (str, optional): Name of watershed.
        transposition_domain_name (str, optional): Name of transposition name.
        **kwargs (Any): Additional keyword arguments.
    """

    def __init__(
        self,
        item_id: str,
        start_datetime: datetime.datetime,
        duration_hours: int,
        watershed: str,
        transposition_domain: str,
        local_directory: str,
        watershed_name: str = None,
        transposition_domain_name: str = None,
        **kwargs: Any,
    ):
        self.item_id = item_id
        self.duration_hours = f"{duration_hours}hrs"
        self.duration = duration_hours
        if not watershed_name:
            self.watershed_name = os.path.basename(watershed).replace(".geojson", "")
        else:
            self.watershed_name = watershed_name

        if not transposition_domain_name:
            self.transposition_domain_name = os.path.basename(transposition_domain).replace(".geojson", "")
        else:
            self.transposition_domain_name = transposition_domain_name

        self.watershed_geometry = watershed
        self.transposition_domain_geometry = transposition_domain
        self.local_directory = local_directory

        self.properties = kwargs.get("properties", {})
        self.stac_extensions = kwargs.get("stac_extensions", None)
        self.href = kwargs.get("href", None)
        self.collection = kwargs.get("collection", None)
        self.extra_fields = kwargs.get("extra_fields", None)
        self.assets = kwargs.get("assets", None)
        self.fiona_env_kwargs = {
            k: v
            for k, v in kwargs.items()
            if k not in ["properties", "stac_extensions", "href", "collection", "extra_fields", "assets"]
        }

        self.start_datetime = start_datetime
        self.end_datetime = start_datetime + self.duration
        super().__init__(
            self.item_id,
            NULL_POLYGON,
            self.transposition_domain_geometry.bounds,
            self.start_datetime,
            self.properties,
            self.start_datetime,
            self.end_datetime,
            self.stac_extensions,
            self.href,
            self.collection,
            self.extra_fields,
            self.assets,
        )

        self._register_extensions()
        self._aorc_source_data: xr.Dataset | None = None
        self._aorc_paths: list[str] | None = None
        self._transpose: Transpose | None = None
        self._sum_aorc: xr.DataArray | None = None
        self._transposed_watershed: Polygon | None = None
        self._transposition_transform: Affine | None = None
        self._stats: dict | None = None

    def _register_extensions(self) -> None:
        """Register item extensions."""
        ProjectionExtension.add_to(self)
        StorageExtension.add_to(self)

    @property
    def aorc_paths(self) -> list[str]:
        """Construct s3 paths for AORC datasets for given start time and duration."""
        if self._aorc_paths is None:
            if self.end_datetime.year == self.start_datetime.year:
                self._aorc_paths = [f"{NOAA_AORC_S3_BASE_URL}/{self.start_datetime.year}.zarr"]
            else:
                year_list = []
                current_year = self.start_datetime.year
                while current_year <= self.end_datetime.year:
                    year_list.append(current_year)
                    current_year += 1
                self._aorc_paths = [f"{NOAA_AORC_S3_BASE_URL}/{year}.zarr" for year in year_list]
            logging.debug("year_list for %s: %s", self.start_datetime, self._aorc_paths)
        return self._aorc_paths

    @property
    def aorc_source_data(self) -> xr.Dataset:
        """Extract AORC source data.

        - reads AORC data into memory as multifile dataset using s3 paths
        - doesn't read the entire ZARR files, instead just reads slice of data corresponding to transposition domain geometry and limited to start and end time
        - adds ZARR files to assets if they don't exist already
        """
        if self._aorc_source_data is None:
            s3_out = s3fs.S3FileSystem(anon=True)
            fileset = [s3fs.S3Map(root=aorc_path, s3=s3_out, check=False) for aorc_path in self.aorc_paths]
            ds = xr.open_mfdataset(fileset, engine="zarr", chunks="auto", consolidated=True)

            transposition_geom_for_clip = self.transposition_domain_geometry
            bounds = transposition_geom_for_clip.bounds
            # adjust start slice to make sure start datetime is exclusive minimum (get data > start not data >= start)
            start_timeslice_value = self.start_datetime + datetime.timedelta(hours=1)
            subsection = ds.sel(
                time=slice(start_timeslice_value, self.end_datetime),
                longitude=slice(bounds[0], bounds[2]),
                latitude=slice(bounds[1], bounds[3]),
            )
            self._aorc_source_data = subsection.rio.clip([transposition_geom_for_clip], drop=True, all_touched=True)
            for aorc_path in self.aorc_paths:
                aorc_year = int(os.path.basename(aorc_path).replace(".zarr", ""))
                aorc_start_datetime = datetime.datetime(
                    year=aorc_year, month=1, day=1, hour=0, tzinfo=datetime.timezone.utc
                )
                aorc_end_datetime = datetime.datetime(
                    year=aorc_year + 1, month=1, day=1, hour=0, tzinfo=datetime.timezone.utc
                )
                asset = Asset(
                    aorc_path,
                    media_type=MediaType.ZARR,
                    extra_fields={
                        "start_datetime": aorc_start_datetime.isoformat(),
                        "end_datetime": aorc_end_datetime.isoformat(),
                    },
                    roles=[MediaType.ZARR],
                )
                storage = StorageExtension.ext(asset)
                storage.platform = CloudPlatform.AWS
                self.add_asset(f"AORC_{aorc_year}", asset)

        return self._aorc_source_data

    @property
    def transpose(self) -> Transpose:
        """Create transpose class to use for transposition functions."""
        if self._transpose is None:
            watershed_geom_for_transpose = self.watershed_geometry
            self._transpose = Transpose(
                self.sum_aorc["APCP_surface"], watershed_geom_for_transpose, AORC_X_VAR, AORC_Y_VAR
            )
        return self._transpose

    @property
    def sum_aorc(self) -> xr.DataArray:
        """Sum AORC precipitation data over the duration."""
        if self._sum_aorc is None:
            self._sum_aorc = self.aorc_source_data.sum(dim="time", skipna=True, min_count=1)
        return self._sum_aorc

    @staticmethod
    def _create_stats(array: np.ndarray) -> dict:
        """Create stats from array."""
        return {
            "min": round(float(np.nanmin(array)) * MM_TO_INCH_CONVERSION_FACTOR, 2),
            "mean": round(float(np.nanmean(array)) * MM_TO_INCH_CONVERSION_FACTOR, 2),
            "max": round(float(np.nanmax(array)) * MM_TO_INCH_CONVERSION_FACTOR, 2),
            "units": "inches",
            # "sum": float(np.nansum(array)) * MM_TO_INCH_CONVERSION_FACTOR,
        }

    def max_transpose(self, add_properties: bool = True) -> tuple[Polygon, Affine, dict]:
        """Get max transpose.

        - convert max array to polygon
        - add stats object to item properties
        - record transpose centroid as item geometry
        - record max shift (as affine transform) to item properties
        - return polygon, transform, and stats
        """
        if not all([self._transposed_watershed, self._transposition_transform, self._stats]):
            self._transposed_watershed, self._transposition_transform, self._stats = self.transpose.max_transpose(
                self._create_stats
            )
        if add_properties:
            self.geometry = json.loads(to_geojson(self._transposed_watershed.centroid))
            self.properties["aorc:statistics"] = self._stats
            self.properties["aorc:transform"] = {
                "a": self._transposition_transform.a,
                "b": self._transposition_transform.b,
                "c": self._transposition_transform.c,
                "d": self._transposition_transform.d,
                "e": self._transposition_transform.e,
                "f": self._transposition_transform.f,
            }
        return (
            self._transposed_watershed,
            self._transposition_transform,
            self._stats,
            self._transposed_watershed.centroid,
        )

    def max_precip_point(self):
        """Add max precipitation location coordinates to item properties."""
        precip_ds = self.sum_aorc["APCP_surface"].compute()
        max_idx = precip_ds.argmax(dim=["latitude", "longitude"])
        lat = precip_ds.latitude[max_idx["latitude"]].item()
        lon = precip_ds.longitude[max_idx["longitude"]].item()
        self.properties["aorc:max_precip_location"] = {"latitude": round(lat, 4), "longitude": round(lon, 4)}

    def aorc_thumbnail(
        self,
        scale_max: float,
        add_asset: bool = True,
        write: bool = True,
        return_fig: bool = False,
    ) -> Figure:
        """Create AORC STAC item thumbnail.

        Creates matplotlib figure showing:
        - location of transposed watershed with maximum precip accumulation
        - original location of watershed
        - valid area of transposition
        - original transposition domain
        """
        if self._transposed_watershed is None:
            self._transposed_watershed, self._transposition_transform, self._stats = self.transpose.max_transpose(
                self._create_stats
            )
        fig, ax = plt.subplots(figsize=(5, 5))
        fig.set_facecolor("w")
        colormap = plt.get_cmap("Spectral_r")
        (self.sum_aorc["APCP_surface"] * MM_TO_INCH_CONVERSION_FACTOR).plot(
            ax=ax, cmap=colormap, cbar_kwargs={"label": "Accumulation (Inches)"}, vmin=0, vmax=scale_max
        )
        valid_area_plt_polygon = patches.Polygon(
            np.column_stack(self.transpose.valid_spaces_polygon.exterior.coords.xy),
            lw=0.7,
            facecolor="none",
            edgecolor="gray",
        )
        ax.add_patch(valid_area_plt_polygon)
        watershed_plt_polygon = patches.Polygon(
            np.column_stack(self.watershed_geometry.exterior.coords.xy),
            lw=0.7,
            facecolor="none",
            edgecolor="gray",
        )
        ax.add_patch(watershed_plt_polygon)
        transposed_watershed_plt_polygon = patches.Polygon(
            np.column_stack(self._transposed_watershed.exterior.coords.xy),
            lw=1,
            facecolor="none",
            edgecolor="black",
        )
        ax.add_patch(transposed_watershed_plt_polygon)
        ax.set(title=None, xlabel=None, ylabel=None)

        if add_asset | write:
            if not os.path.isdir(self.local_directory):
                os.makedirs(self.local_directory)

            filename = f"{self.item_id}.thumbnail.png"
            fn = os.path.join(self.local_directory, filename)
            fig.savefig(fn, bbox_inches="tight")
            asset = Asset(filename, media_type=MediaType.PNG, roles=["thumbnail"])
            self.add_asset("thumbnail", asset)
        if return_fig:
            return fig
        else:
            plt.close()


def valid_spaces_item(watershed: Item, transposition_region: Item, storm_duration: int = 72) -> Polygon:
    """Search a sample zarr dataset to identify valid spaces for transposition. datetime.datetime(1980, 5, 1) is used as a start time for the search."""
    s3 = s3fs.S3FileSystem(anon=True)
    start_time = datetime.datetime(1980, 5, 1)
    sample_data = s3fs.S3Map(root=f"{NOAA_AORC_S3_BASE_URL}/{start_time.year}.zarr", s3=s3)
    ds = xr.open_dataset(sample_data, engine="zarr", chunks="auto", consolidated=True)
    bounds = shape(transposition_region.geometry).bounds

    subset = ds.sel(
        time=slice(start_time, start_time + datetime.timedelta(hours=storm_duration)),
        longitude=slice(bounds[0], bounds[2]),
        latitude=slice(bounds[1], bounds[3]),
    )

    clipped_data = subset.rio.clip([shape(transposition_region.geometry)], drop=True, all_touched=True)
    transpose = Transpose(
        clipped_data[AORC_PRECIP_VARIABLE].sum(dim="time", skipna=True, min_count=1),
        shape(watershed.geometry),
        AORC_X_VAR,
        AORC_Y_VAR,
    )
    return transpose.valid_spaces_polygon



"""Logging utility and setup."""

import logging
import sys

SUPPRESS_LOGS = ["boto3", "botocore", "geopandas", "fiona", "rasterio", "pyogrio", "xarray", "shapely"]


def initialize_logger(json_logging: bool = False, level: int = logging.INFO):
    """Initialize the logger."""
    datefmt = "%Y-%m-%dT%H:%M:%SZ"
    if json_logging:
        for module in SUPPRESS_LOGS:
            logging.getLogger(module).setLevel(logging.WARNING)

        class FlushStreamHandler(logging.StreamHandler):
            def emit(self, record):
                super().emit(record)
                self.flush()

        handler = FlushStreamHandler(sys.stdout)

        logging.basicConfig(
            level=level,
            handlers=[handler],
            format="""{"time": "%(asctime)s" , "level": "%(levelname)s", "msg": "%(message)s"}""",
            datefmt=datefmt,
        )
    else:
        for package in SUPPRESS_LOGS:
            logging.getLogger(package).setLevel(logging.ERROR)
        logging.basicConfig(level=level, format="%(asctime)s | %(levelname)s | %(message)s", datefmt=datefmt)
    # boto3.set_stream_logger(name="botocore.credentials", level=logging.ERROR)



"""Script to extract metadata from a directory of local dss files including the first precipitation and temperature path names the rank, and the stormtype."""

import os
import json
import re
from hecdss import HecDss
from datetime import datetime


def parse_d_part(d_part):
    """Convert D part of dss path to datetime."""
    try:
        return datetime.strptime(d_part, "%d%b%Y:%H%M")
    except Exception:
        return None


def extract_storm_type_and_rank(filename):
    """Strip the storm type and rank given a dss path in format yyymmdd_72hr_st(x)_r(xxx).dss ."""
    match = re.search(r"_([^_]+)_(r\d+)\.dss$", filename)
    storm_type = match.group(1)
    rank = match.group(2)
    return storm_type, rank


def generate_dss_metadata(dss_dir, output_path):
    """
    Generate metadata from DSS files including earliest precipitation and temperature paths.

    Args:
        dss_dir (str): Directory containing DSS files.
        output_path (str): Path to save the output JSON file.
    """
    result = {}

    for filename in os.listdir(dss_dir):
        if not filename.endswith(".dss"):
            continue

        filepath = os.path.join(dss_dir, filename)
        precip_path = None
        temp_path = None
        earliest_precip_time = None
        earliest_temp_time = None

        try:
            with HecDss(filepath) as dss:
                for path_obj in dss.get_catalog():
                    path = str(path_obj).strip("/")
                    parts = path.split("/")
                    if len(parts) < 6:
                        continue

                    part_c = parts[2].upper()
                    part_d = parts[3]
                    dt = parse_d_part(part_d)

                    if not dt:
                        continue

                    if part_c == "PRECIPITATION":
                        if precip_path is None or dt < earliest_precip_time:
                            precip_path = path_obj
                            earliest_precip_time = dt
                    elif part_c == "TEMPERATURE":
                        if temp_path is None or dt < earliest_temp_time:
                            temp_path = path_obj
                            earliest_temp_time = dt
        except Exception as e:
            print(f"ERROR reading {filename}: {e}")
            continue

        storm_type, rank = extract_storm_type_and_rank(filename)
        result[filename] = {
            "precip_pathname": str(precip_path) if precip_path else None,
            "temperature_pathname": str(temp_path) if temp_path else None,
            "storm_type": storm_type,
            "rank": rank,
        }

    with open(output_path, "w") as f:
        json.dump(result, f, indent=2)

    print(f"Saved pathnames to {output_path}")


if __name__ == "__main__":
    dss_dir = "<path_to>/dss_downloads"  # directory of local dss files
    output_path = "dss_metadata.json"

    generate_dss_metadata(dss_dir, output_path)



"""Utility functions for stormhub."""

import json
import logging
import socket
from datetime import datetime, timedelta
from typing import List
import os

from pystac import Link, Collection
from shapely.geometry import mapping, shape

STORMHUB_REF_LINK = Link(
    rel="Processing",
    target="https://github.com/Dewberry/stormhub",
    title="Source Code",
    media_type="text/html",
    extra_fields={"Description": "Source code used to generate STAC objects"},
)


def is_port_in_use(port: int = 8080, host: str = "http://localhost") -> bool:
    """Check if a given port is already in use."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind((host, port))
            return False
        except socket.error:
            return True


def load_config(config_file: str) -> dict:
    """Load a json config file."""
    with open(config_file, "r", encoding="utf-8") as f:
        return json.load(f)


def validate_config(config: dict) -> dict:
    """Validate a config dictionary against required keys."""
    required_keys = {
        "watershed": ["id", "geometry_file", "description"],
        "transposition_region": ["id", "geometry_file", "description"],
    }

    for key, sub_keys in required_keys.items():
        if key not in config:
            raise ValueError(f"Missing required section: {key}")
        for sub_key in sub_keys:
            if sub_key not in config[key] or not config[key][sub_key]:
                raise ValueError(f"Missing value for {sub_key} in section {key}")
    return config


def generate_date_range(
    start_date: str, end_date: str, every_n_hours: int = 6, date_format: str = "%Y-%m-%d"
) -> List[datetime]:
    """Generate a list of datetime objects at a given interval between start and end dates."""
    start = datetime.strptime(start_date, date_format)
    end = datetime.strptime(end_date, date_format)

    date_range = []
    current_date = start
    while current_date <= end:
        date_range.append(current_date)
        current_date += timedelta(hours=every_n_hours)

    return date_range


def create_feature_collection_from_items(
    collection: Collection, output_geojson: str, select_properties: str = "aorc:statistics"
):
    """Generate a geojson feature collection from a collection of STAC items."""
    features = []
    for item in collection.get_all_items():
        geom = shape(item.geometry)
        if geom.is_empty:
            continue

        feature = {
            "type": "Feature",
            "geometry": mapping(geom),
            "properties": {
                "id": item.id,
                select_properties: item.properties.get(select_properties),
                # **item.properties,
            },
        }
        features.append(feature)

    feature_collection = {"type": "FeatureCollection", "features": features}

    with open(output_geojson, "w", encoding="utf-8") as f:
        json.dump(feature_collection, f, indent=4)

    logging.info("FeatureCollection saved to %s", output_geojson)


class StacPathManager:
    """Build consistent paths for STAC items and collections assuming a top-level local catalog."""

    def __init__(self, local_catalog_dir: str):
        self._catalog_dir = os.path.abspath(local_catalog_dir)

    @property
    def catalog_dir(self):
        """Build Catalog directory path."""
        return self._catalog_dir

    @property
    def catalog_file(self):
        """Build Catalog file path."""
        return os.path.join(self._catalog_dir, "catalog.json")

    def storm_collection_id(self, duration: int) -> str:
        """Build storm collection id."""
        return f"{duration}hr-events"

    def catalog_item(self, item_id: str) -> str:
        """Build Catalog item path."""
        return os.path.join(self.catalog_dir, item_id, f"{item_id}.json")

    def catalog_asset(self, item_id: str, asset_dir: str = "hydro_domains") -> str:
        """Build Catalog asset path."""
        return os.path.join(self.catalog_dir, asset_dir, f"{item_id}.json")

    def collection_file(self, collection_id: str) -> str:
        """Build Collection file path."""
        return os.path.join(self.catalog_dir, collection_id, "collection.json")

    def collection_dir(self, collection_id: str) -> str:
        """Build Collection directory path."""
        return os.path.join(self.catalog_dir, collection_id)

    def collection_asset(self, collection_id: str, filename: str) -> str:
        """Build Collection asset path."""
        return os.path.join(self.catalog_dir, collection_id, filename)

    def collection_item_dir(self, collection_id: str, item_id: str) -> str:
        """Build Collection item directory path."""
        return os.path.join(self.catalog_dir, collection_id, item_id)

    def collection_item(self, collection_id: str, item_id: str) -> str:
        """Build Collection item path."""
        return os.path.join(self.catalog_dir, collection_id, item_id, f"{item_id}.json")

    def collection_item_asset(self, collection_id: str, item_id: str, filename: str) -> str:
        """Build Collection item asset path."""
        return os.path.join(self.catalog_dir, collection_id, item_id, filename)


def file_table(data: dict, col1: str, col2: str):
    """Convert a dictionary into a list of dictionaries, suitable for creating a table."""
    table = []
    for k, v in data.items():
        table.append({col1: k, col2: v})
    return table



"""Constants."""

NOAA_AORC_S3_BASE_URL = "s3://noaa-nws-aorc-v1-1-1km"

AORC_PRECIP_VARIABLE = "APCP_surface"
AORC_X_VAR = "longitude"
AORC_Y_VAR = "latitude"

MM_TO_INCH_CONVERSION_FACTOR = 0.03937007874015748

SHG_WKT = 'PROJCS["USA_Contiguous_Albers_Equal_Area_Conic_USGS_version",GEOGCS["GCS_North_American_1983",DATUM["D_North_American_1983",SPHEROID["GRS_1980",6378137.0,298.257222101]],PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]],PROJECTION["Albers"],PARAMETER["False_Easting",0.0],PARAMETER["False_Northing",0.0],PARAMETER["Central_Meridian",-96.0],PARAMETER["Standard_Parallel_1",29.5],PARAMETER["Standard_Parallel_2",45.5],PARAMETER["Latitude_Of_Origin",23.0],UNIT["Meter",1.0]]'
"""CRS definition, based on EPSG:5070"""

KM_TO_M_CONVERSION_FACTOR = 1000



"""Module for createing USGS STAC objects."""

import logging
import os
from datetime import datetime
from pathlib import Path
from typing import Optional, List

import geopandas as gpd
from shapely.geometry import shape
from dataretrieval import nwis, NoSitesError
from pystac import Asset, Item, MediaType, RelType, Link
import pystac
from stormhub.hydro.utils import log_pearson_iii


from stormhub.hydro.plots import (
    plot_ams,
    plot_ams_seasonal,
    plot_log_pearson_iii,
    plot_nwis_statistics,
)
from stormhub.utils import file_table


class UsgsGage(Item):
    """A class representing a USGS gage as a STAC item."""

    @classmethod
    def from_usgs(cls, gage_number: str, href: Optional[str] = None, **kwargs):
        """Create a STAC Item representing a USGS stream gage.

        Parameters
        ----------
            gage_number (str): USGS gage number.
            href (Optional[str]): Item href for the created USGS gage item. Optional

        Returns
        -------
            pystac.Item: A STAC Item representing the USGS gage.
        """
        if href is None:
            href = f"{gage_number}.json"

        site_data = cls._load_site_data(gage_number)

        geometry = {"type": "Point", "coordinates": [site_data["dec_long_va"], site_data["dec_lat_va"]]}
        bbox = [
            site_data["dec_long_va"],
            site_data["dec_lat_va"],
            site_data["dec_long_va"],
            site_data["dec_lat_va"],
        ]
        properties = {
            "site_no": site_data["site_no"],
            "station_nm": site_data["station_nm"],
            "huc_cd": str(site_data["huc_cd"]),
            "drain_area_va": float(site_data["drain_area_va"]),
            "daily_values": {
                "begin_date": site_data["daily_values"]["begin_date"],
                "end_date": site_data["daily_values"]["end_date"],
            },
            "site_retrieved": site_data["site_retrieved"],
        }
        start_datetime, end_datetime = cls.start_end_dates(gage_number)
        if properties["daily_values"]["begin_date"] is None and properties["daily_values"]["end_date"] is None:
            properties["daily_values"]["end_date"] = end_datetime.strftime("%Y-%m-%d") if end_datetime else None
            properties["daily_values"]["begin_date"] = start_datetime.strftime("%Y-%m-%d") if start_datetime else None

        logging.info(f"Creating UsgsGage {gage_number} {site_data['station_nm']}")

        usgs_gage = cls(
            id=gage_number,
            geometry=geometry,
            bbox=bbox,
            datetime=datetime.now(),
            properties=properties,
            start_datetime=start_datetime,
            end_datetime=end_datetime,
            href=href,
            **kwargs,
        )

        gage_url = f"https://waterdata.usgs.gov/nwis/inventory/?site_no={properties['site_no']}"
        usgs_gage.add_link(Link(rel=RelType.VIA, target=gage_url, title="USGS NWIS Site Information"))
        return usgs_gage

    def __repr__(self):
        """Return string representation of the UsgsGage object."""
        return f"<UsgsGage {self.id} {self.properties['station_nm']}>"

    @staticmethod
    def _load_site_data(gage_number: str) -> dict:
        """Query NWIS for site information."""
        resp = nwis.get_record(sites=gage_number, service="site")

        return {
            "site_no": resp["site_no"].iloc[0],
            "station_nm": resp["station_nm"].iloc[0],
            "dec_lat_va": float(resp["dec_lat_va"].iloc[0]),
            "dec_long_va": float(resp["dec_long_va"].iloc[0]),
            "drain_area_va": resp["drain_area_va"].iloc[0],
            "huc_cd": resp["huc_cd"].iloc[0],
            "alt_datum_cd": resp["alt_datum_cd"].iloc[0],
            "site_retrieved": datetime.now().isoformat(),
            "daily_values": {
                "begin_date": resp["begin_date"].iloc[0] if "begin_date" in resp else None,
                "end_date": resp["end_date"].iloc[0] if "end_date" in resp else None,
            },
        }

    def start_end_dates(gage_id: str):
        """Retrieve start and end dates from oldest and newest daily value records."""
        startDate = "1900-01-01"
        endDate = datetime.now().strftime("%Y-%m-%d")
        dv = nwis.get_dv(gage_id, startDate, endDate)[0]
        dv_sorted = dv.sort_index()
        if len(dv_sorted) > 1:
            return dv_sorted.index.min().to_pydatetime(), dv_sorted.index.max().to_pydatetime()
        else:
            return None, None

    def get_peaks(self, item_dir: str, make_plots: bool = True):
        """Retrieve annual maximum series from NWIS and make the plots associated with those values."""
        gage_id = self.properties["site_no"]

        try:
            df = nwis.get_record(service="peaks", sites=[gage_id])
        except NoSitesError:
            logging.warning(f"Peaks could not be found for gage id: {gage_id}")
            return

        file_name = os.path.join(item_dir, f"{gage_id}-ams.pq")

        if not os.path.exists(item_dir):
            os.makedirs(item_dir)
        df.to_parquet(file_name)

        try:
            peaks = log_pearson_iii(df["peak_va"])
        except ValueError:
            logging.warning(f"LP3 peaks stats could not be calculated for gage id: {gage_id}")
            return

        asset = Asset(
            file_name,
            media_type=MediaType.PARQUET,
            roles=["data"],
            extra_fields={"file:values": file_table(peaks, "return_period", "discharge_CFS_(Approximate)")},
        )

        self.add_asset("annual_maxima_series", asset)

        if make_plots:
            # AMS Plot 1
            filename = os.path.join(item_dir, f"{gage_id}-ams.png")
            plot_ams(df, gage_id, filename)

            asset = Asset(filename, media_type=MediaType.PNG, roles=["thumbnail"])
            self.add_asset("ams_plot", asset)

            # AMS Plot 2
            filename = os.path.join(item_dir, f"{gage_id}-ams-seasonal.png")
            plot_ams_seasonal(df, gage_id, filename)

            asset = Asset(filename, media_type=MediaType.PNG, roles=["thumbnail"])
            self.add_asset("ams_seasons_plot", asset)

            # LPII Plot
            filename = os.path.join(item_dir, f"{gage_id}-ams-lpiii.png")
            plot_log_pearson_iii(df["peak_va"], gage_id, filename)

            asset = Asset(filename, media_type=MediaType.PNG, roles=["thumbnail"])
            self.add_asset("ams_LPIII_plot", asset)

    def get_flow_stats(self, item_dir: str, make_plots: bool = True):
        """Retrieve and plot day of the year flow statistics."""
        gage_id = self.properties["site_no"]

        try:
            df = nwis.get_stats(sites=gage_id, parameterCd="00060")[0]
        except IndexError:
            logging.warning(f"Flow stats could not be found for gage_id: {gage_id}")
            return

        file_name = os.path.join(item_dir, f"{gage_id}-flow-stats.pq")

        if not os.path.exists(item_dir):
            os.makedirs(item_dir)

        df.to_parquet(file_name)

        asset = Asset(file_name, media_type=MediaType.PARQUET, roles=["data"])

        self.add_asset("flow_stats", asset)

        if make_plots:
            # AMS Plot 1
            filename = os.path.join(item_dir, f"{gage_id}-flow-stats.png")
            plot_nwis_statistics(df, gage_id, filename)

            asset = Asset(filename, media_type=MediaType.PNG, roles=["thumbnail"])
            self.add_asset("flow_statistics_plot", asset)


def from_stac(href: str) -> UsgsGage:
    """Create a UsgsGage from a STAC Item."""
    return UsgsGage.from_file(href)


class GageCollection(pystac.Collection):
    """USGS gage collection."""

    def __init__(self, collection_id: str, items: List[pystac.Item], href):
        """
        Initialize a GageCollection instance.

        Parameters
        ----------
            collection_id (str): The ID of the collection.
            items (List[pystac.Item]): List of STAC items to include in the collection.
        """
        spatial_extents = [item.bbox for item in items if item.bbox]
        temporal_extents = [item.datetime for item in items if item.datetime is not None]

        collection_extent = pystac.Extent(
            spatial=pystac.SpatialExtent(
                bboxes=[
                    [
                        min(b[0] for b in spatial_extents),
                        min(b[1] for b in spatial_extents),
                        max(b[2] for b in spatial_extents),
                        max(b[3] for b in spatial_extents),
                    ]
                ]
            ),
            temporal=pystac.TemporalExtent(intervals=[[min(temporal_extents), max(temporal_extents)]]),
        )

        super().__init__(
            id=collection_id,
            description="STAC collection generated from gage items.",
            extent=collection_extent,
            href=href,
        )

        for item in items:
            self.add_item_to_collection(item)

    def add_item_to_collection(self, item: Item, override: bool = False):
        """
        Add an item to the collection.

        Parameters
        ----------
            item (Item): The STAC item to add.
            override (bool): Whether to override an existing item with the same ID.
        """
        existing_ids = {item.id for item in self.get_all_items()}

        if item.id in existing_ids:
            if override:
                self.remove_item(item.id)
                item.set_parent(self)
                self.add_item(item)
                logging.info(f"Overwriting (existing) item with ID '{item.id}'.")
            else:
                logging.error(
                    f"Item with ID '{item.id}' already exists in the collection. Use `override=True` to overwrite."
                )
        else:
            item.set_parent(self)
            self.add_item(item)
            logging.info(f"Added item with ID '{item.id}' to the collection.")

    def items_to_geojson(self, items: List[pystac.Item], geojson_dir: str):
        """Add a list of STAC items to a geojson and saves it as a collection asset."""
        records = []
        for item in items:
            geom = shape(item.geometry)
            records.append({"site_no": item.properties.get("site_no"), "geometry": geom})

        gdf = gpd.GeoDataFrame(records, crs="EPSG:4326")
        geojson_path = geojson_dir.joinpath("gages.geojson")
        gdf.to_file(geojson_path, driver="GeoJSON")

        geojson_asset = pystac.Asset(
            href=str(geojson_path.relative_to(geojson_dir)).replace("\\", "/"),
            media_type=pystac.MediaType.GEOJSON,
            title="Gages GeoJSON",
        )
        self.add_asset("geojson", geojson_asset)


def new_gage_catalog(catalog_id: str, local_directory: str, catalog_description: str) -> pystac.Catalog:
    """
    Create a new STAC catalog for storing USGS gage collection.

    Parameters
    ----------
        catalog_id (str): Unique id for the STAC catalog.
        local_directory (Optional[str]): Directory where the catalog will be saved.
        catalog_description (str): The description of the catalog.

    Returns
    -------
        pystac.Catalog: The created STAC catalog.
    """
    if not local_directory:
        local_directory = os.getcwd()

    catalog = pystac.Catalog(id=catalog_id, description=catalog_description)
    catalog.normalize_and_save(root_href=local_directory, catalog_type=pystac.CatalogType.SELF_CONTAINED)
    return catalog


def new_gage_collection(catalog: pystac.Catalog, gage_numbers: List[str], directory: str) -> None:
    """
    Create a new STAC collection for USGS gages and adds it to an existing catalog.

    Parameters
    ----------
        catalog (pystac.Catalog): The STAC catalog which the collection will be added to.
        gage_numbers (List[str]): A list of USGS gage site numbers to add to the collection.
        directory (str): The directory where the STAC collection and items will be stored.
    """
    base_dir = Path(directory)
    gages_dir = base_dir.joinpath("gages")
    gages_dir.mkdir(parents=True, exist_ok=True)
    collection_href = base_dir.joinpath("collection.json")

    items = []
    for gage_number in gage_numbers:
        try:
            gage_item_dir = gages_dir.joinpath(gage_number)
            gage_item_dir.mkdir(parents=True, exist_ok=True)

            gage = UsgsGage.from_usgs(gage_number, href=str(gage_item_dir.joinpath(f"{gage_number}.json")))
            gage.get_flow_stats(str(gage_item_dir))
            gage.get_peaks(str(gage_item_dir))

            for asset in gage.assets.values():
                asset.href = os.path.relpath(asset.href, gage_item_dir).replace("\\", "/")

            gage.save_object()
            items.append(gage)
        except Exception as e:
            logging.error(f"Gage {gage_number} failed eith exception: {e}")

    collection = GageCollection("gages", items, str(collection_href))
    collection.items_to_geojson(items, gages_dir)

    catalog.add_child(collection)
    catalog.normalize_and_save(root_href=str(base_dir), catalog_type=pystac.CatalogType.SELF_CONTAINED)



"""USGS gage utility functions."""

import logging
from typing import List, Optional

import geopandas as gpd
import numpy as np
import pandas as pd
import scipy.stats as stats
from dataretrieval import NoSitesError, nwis


def find_gages_in_watershed(watershed: str, min_num_records: Optional[int] = None) -> List[str]:
    """
    Identify USGS gages within a given watershed and optionally filters by a minimum number of records.

    Parameters
    ----------
        watershed (str): Path to a GeoJSON containing the watershed geometry.
        min_num_records (Optional[int]): Minimum number of records required for a gage to be included. If None, all valid gages within the watershed are returned.

    Returns
    -------
        List[str]: A list of USGS gage site numbers that meet the criteria.
    """
    logging.info("Searching watershed for gages")
    watershed = gpd.read_file(watershed)
    bbox = [round(coord, 6) for coord in watershed.total_bounds.tolist()]

    gages_in_watershed_bbox = nwis.get_info(bBox=bbox, parameterCd=["00060", "00065"])[0]

    watershed_geom = watershed.iloc[0].geometry
    gages_within_watershed = gages_in_watershed_bbox[gages_in_watershed_bbox.within(watershed_geom)]

    candidate_gages = []
    logging.info(
        f"USGS API returned {len(gages_within_watershed)} responses, filtering for valid gages based in site_no length (must = 8 characters)"
    )
    for row in gages_within_watershed.itertuples():
        site_id = row.site_no
        logging.info(f"Checking gage {site_id}")
        if len(site_id) == 8:
            candidate_gages.append(site_id)
        else:
            logging.info(f"Removing potentially invalid gage {site_id}")

    if len(candidate_gages) == 0:
        logging.warning("No valid gages found within given watershed.")
        return []
    else:
        logging.info(f"Found {len(candidate_gages)} valid gage numbers in watershed")

    if min_num_records is None:
        return candidate_gages

    valid_gage_nums = []
    for gage_num in candidate_gages:
        try:
            logging.info(f"Checking period of record for `{gage_num}`")
            if len(nwis.get_record(service="peaks", sites=[gage_num])) >= min_num_records:
                valid_gage_nums.append(gage_num)
        except NoSitesError:
            continue
    logging.info(f"Found {len(valid_gage_nums)} valid gage numbers in watershed with min period of record.")
    return valid_gage_nums


def log_pearson_iii(peak_flows: pd.Series, standard_return_periods: list = [2, 5, 10, 25, 50, 100, 500]) -> dict:
    """Calculate peak flow estimates for specified return periods using the Log-Pearson Type III distribution.

    Args:
        peak_flows (pd.Series): A pandas Series containing peak flow values.
        standard_return_periods (list, optional): A list of return periods for which to calculate peak flow estimates.

    Returns
    -------
        dict: A dictionary where keys are return periods and values are the peak flow estimates.
    """
    log_flows = np.log10(peak_flows.values)
    mean_log = np.mean(log_flows)
    std_log = np.std(log_flows, ddof=1)
    skew_log = stats.skew(log_flows)

    return {
        str(rp): int(10 ** (mean_log + stats.pearson3.ppf(1 - 1 / rp, skew_log) * std_log))
        for rp in standard_return_periods
    }



"""Testing functions."""

import unittest
from math import floor

import fiona
import numpy as np
import rioxarray
import xarray as xr
from affine import Affine
from shapely import Polygon

from stormhub.met.transpose import Transpose


def shapely_polygon_to_geojson(polygon: Polygon) -> dict:
    """
    Convert a Shapely Polygon to GeoJSON format.
    """
    return {
        "type": "Feature",
        "geometry": {
            "type": "Polygon",
            "coordinates": [list(polygon.exterior.coords)],
        },
        "properties": {"id": 1},
    }


def save_polygon_to_geojson(polygon: Polygon, filename: str):
    """
    Save a Shapely Polygon to a GeoJSON file.
    """
    schema = {
        "geometry": "Polygon",
        "properties": {"id": "int"},
    }
    geojson = shapely_polygon_to_geojson(polygon)
    with fiona.open(filename, "w", driver="GeoJSON", crs="EPSG:4326", schema=schema) as dst:
        dst.write(geojson)


def save_xarray_to_raster(data_array: xr.DataArray, filename: str):
    """
    Save an xarray DataArray to a raster file.
    """
    data_array.rio.to_raster(filename, driver="GTiff", dtype="int16")


def create_test_transposition_domain_polygon() -> Polygon:
    """
    Create a test polygon for the transpose function.
    """
    # Define the coordinates of the polygon
    coords = [(3, 0), (5, 2), (4, 4), (0, 5)]

    # Create a Shapely Polygon object
    polygon = Polygon(coords)

    return polygon


def create_test_watershed_polygon() -> Polygon:
    """
    Create a test polygon for the transpose function.
    """
    # Define the coordinates of the polygon
    coords = [(3, 1), (3, 3), (1, 4), (2, 3), (2, 3)]

    # Create a Shapely Polygon object
    polygon = Polygon(coords)

    return polygon


def create_test_data_array(transposition_domain_polygon: Polygon, res: float) -> xr.DataArray:
    """
    Create a test DataArray for the transpose function.
    """
    # Create a 5x5 grid with values ranging from 0 to 24
    min_x, min_y, max_x, max_y = transposition_domain_polygon.bounds
    nx = floor((max_x - min_x) / res)
    ny = floor((max_y - min_y) / res)
    longitudes, latitudes = (
        np.linspace(min_x + res / 2, max_x - res / 2, nx),
        np.linspace(min_y + res / 2, max_y - res / 2, ny),
    )

    # Create sample data
    data = np.arange(nx * ny, dtype=np.float64)
    data = data.reshape((ny, nx))

    # Create the xarray DataArray
    da = xr.DataArray(
        data=data,
        dims=["latitude", "longitude"],
        coords={"latitude": latitudes, "longitude": longitudes},
    )

    rio_da = da.rio.set_spatial_dims(x_dim="longitude", y_dim="latitude")
    rio_da.rio.write_crs("EPSG:4326", inplace=True)
    rio_da = rio_da.rio.clip([transposition_domain_polygon], all_touched=True, drop=True)

    return rio_da


class TestTransposeFunction(unittest.TestCase):
    def setUp(self):
        self.watershed = create_test_watershed_polygon()
        self.transposition_domain = create_test_transposition_domain_polygon()
        self.data_array = create_test_data_array(self.transposition_domain, 1)
        self.transpose = Transpose(self.data_array, self.watershed, "longitude", "latitude")

    def test_valid_spaces_polygon(self):
        """
        Test the valid spaces polygon.
        """
        # Create a test polygon
        test_polygon = Polygon(
            [(0, 5), (4, 5), (4, 4), (5, 4), (5, 1), (4, 1), (4, 0), (2, 0), (2, 1), (1, 1), (1, 3), (0, 3)]
        )

        # Check if the valid spaces polygon is correct
        self.assertTrue(self.transpose.valid_spaces_polygon.equals(test_polygon))

    def test_max_transpose_polygon(self):
        """
        Test the max transpose function.
        """
        # Create a test polygon
        test_polygon = Polygon([(3, 5), (4, 5), (4, 4), (5, 4), (5, 1), (4, 1), (4, 3), (3, 3)])

        # Check if the max transpose polygon is correct
        max_polygon = self.transpose.max_transpose()[0]
        self.assertTrue(max_polygon.equals(test_polygon))

    def test_max_transpose_max_value(self):
        # Create a test max value
        test_max_value = 23

        # Check if the max value is correct
        max_value = self.transpose.max_transpose(np.max)[2]
        self.assertEqual(max_value, test_max_value)

    def test_max_transpose_affine(self):
        # Create a test affine transformation
        test_affine = Affine.translation(2, 0)

        # Check if the affine transformation is correct
        max_affine = self.transpose.max_transpose()[1]
        self.assertEqual(max_affine, test_affine)


if __name__ == "__main__":
    unittest.main()



"""Creates dss files using multithreading for efficiency from a given watershed for each event in a metadata json formatted in the following way.

{
  "1988-09-13": {
    "event_id": "236",
    "st_number": "st5"
  },
  "1993-11-14": {
    "event_id": "190",
    "st_number": "st5"
  }
}.
"""

import os
import json
import logging
from datetime import datetime

from concurrent.futures import ProcessPoolExecutor, as_completed
from stormhub.met.zarr_to_dss import noaa_zarr_to_dss, NOAADataVariable


def setup_logging():
    """Ensure each process sets up its own logging to file."""
    logger = logging.getLogger()
    if not logger.handlers:
        logging.basicConfig(
            filename="storm_processing.log",
            level=logging.INFO,
            format="%(asctime)s - %(levelname)s - %(message)s",
        )


def process_storm(date, attrs, dss_dir, aoi_path, aoi_name):
    """Process each storm event and generate DSS files."""
    setup_logging()

    storm_start_datetime = datetime.strptime(date, "%Y-%m-%d")
    storm_start_str = datetime.strftime(storm_start_datetime, "%Y%m%d")
    output_name = f"{storm_start_str}_72h_{attrs['st_number']}_r{attrs['event_id']}.dss"
    output_dss_path = os.path.join(dss_dir, output_name)

    try:
        if os.path.exists(output_dss_path):
            logging.info(f"SKIPPED {date} - {output_name} already exists.")
            return

        noaa_zarr_to_dss(
            output_dss_path=output_dss_path,
            aoi_geometry_gpkg_path=aoi_path,
            aoi_name=aoi_name,
            storm_start=storm_start_datetime,
            variable_duration_map={
                NOAADataVariable.TMP: 864,
                NOAADataVariable.APCP: 72,
            },
        )

        logging.info(f"SUCCESS {date} - {output_name}")

    except Exception as e:
        logging.error(f"FAILED {date} - {output_name} - Error: {e}")


def main(json_path, dss_dir, aoi_path, aoi_name):
    """Create dss files in parallel."""
    with open(json_path) as f:
        storms = json.load(f)

    os.makedirs(dss_dir, exist_ok=True)

    with ProcessPoolExecutor(max_workers=2) as executor:
        futures = [
            executor.submit(process_storm, date, attrs, dss_dir, aoi_path, aoi_name) for date, attrs in storms.items()
        ]

        for future in as_completed(futures):
            future.result()


if __name__ == "__main__":
    json_path = "storm_events.json"
    dss_dir = "dss_outputs"
    aoi_path = "trinity-transpo-area-v01.geojson"
    aoi_name = "TRINITY"
    main(json_path, dss_dir, aoi_path, aoi_name)



"""Updates the metadata json created by extract_dss_metadata.py to include storm center coordinates from stac items by matching the dss rank to the stac item prefix and id."""

import os
import re
import json
import boto3
from io import BytesIO
from dotenv import load_dotenv
import logging
from stormhub.logger import initialize_logger

initialize_logger()
load_dotenv()
s3 = boto3.client("s3")


def extract_rank(filename):
    """Extract rank from filename, removing leading 0's ('_r003' -> '3')."""
    match = re.search(r"_r(\d+)", filename)
    return str(int(match.group(1))) if match else None


def update_dss_metadata_with_coords(json_path, dss_filenames, bucket, stac_prefix):
    """Add storm center coordinates to DSS metadata based on matched STAC items."""
    for dss_filename in dss_filenames:
        rank = extract_rank(dss_filename)
        if not rank:
            logging.error(f"Skipping {dss_filename}: no rank found.")
            continue

        stac_key = f"{stac_prefix}{rank}/{rank}.json"
        logging.info(f"Extracting coordinates from: {stac_key}")
        try:
            obj = s3.get_object(Bucket=bucket, Key=stac_key)
            item = json.load(obj["Body"])
            transform = item.get("properties", {}).get("aorc:transform", {})
            dss_filenames[dss_filename]["storm_center_lon"] = round(transform.get("storm_center_lon", 0), 4)
            dss_filenames[dss_filename]["storm_center_lat"] = round(transform.get("storm_center_lat", 0), 4)
        except Exception as e:
            logging.error(f"Failed to update {dss_filename}: {e}")

    with open(json_path, "w") as f:
        json.dump(dss_filenames, f, indent=2)

    logging.info(f"Updated storm center coordinates in {json_path}.")


def main(json_path, bucket_name, stac_prefix):
    """Add STAC Item coordinates to metadata json."""
    with open(json_path, "r") as f:
        dss_filenames = json.load(f)
    update_dss_metadata_with_coords(json_path, dss_filenames, bucket_name, stac_prefix)


if __name__ == "__main__":
    bucket_name = "trinity-pilot"
    stac_prefix = "stac/prod-support/storms/72hr-events/"
    json_path = "<path_to>/dss_metadata.json"

    main(json_path, bucket_name, stac_prefix)



"""Class to handle transpoition functionality."""

from typing import Any, Callable

import numpy as np
import xarray as xr
from affine import Affine
from rasterio.features import shapes
from rasterio.mask import geometry_mask
from rasterio.windows import Window, get_data_window
from shapely import Polygon
from shapely.affinity import translate
from shapely.geometry import shape


class Transpose:
    """
    A class to handle the transposition of a watershed mask over a data array.

    Attributes
    ----------
        data_array (xr.DataArray): The data array to be transposed.
        watershed (Polygon): The watershed polygon.
        x_var (str): The x variable name in the data array.
        y_var (str): The y variable name in the data array.
    """

    def __init__(self, data_array: xr.DataArray, watershed: Polygon, x_var: str, y_var: str) -> None:
        """
        Initialize the Transpose class.

        Args:
            data_array (xr.DataArray): The data array to be transposed.
            watershed (Polygon): The watershed polygon.
            x_var (str): The x variable name in the data array.
            y_var (str): The y variable name in the data array.
        """
        self.data_array = data_array
        self.watershed = watershed
        self.x_var = x_var
        self.y_var = y_var
        self.x_cellsize, self.y_cellsize = self.data_array.rio.resolution()
        self.transform = self.data_array.rio.transform()
        self.width = self.data_array.rio.width
        self.height = self.data_array.rio.height
        self._np_data_array = None
        self._watershed_window = None
        self._watershed_mask = None
        self._watershed_mask_clipped = None
        self._valid_shifts = None
        self._valid_spaces = None
        self._valid_spaces_polygon = None
        self._data_array_x_coords = None
        self._data_array_y_coords = None

    def _calculate_watershed_mask_and_window(self) -> None:
        """
        Calculate the watershed mask and window.

        This method creates a mask array and determines the window of the watershed
        within the data array.
        """
        mask_array = geometry_mask(
            [self.watershed],
            out_shape=(int(self.data_array.rio.height), int(self.data_array.rio.width)),
            transform=self.data_array.rio.transform(recalc=True),
            all_touched=True,
            invert=True,
        )
        self._watershed_mask = mask_array
        window = get_data_window(np.ma.masked_array(mask_array, ~mask_array))
        self._watershed_window = window
        (row_start, row_stop), (col_start, col_stop) = window.toranges()
        self._watershed_mask_clipped = mask_array[row_start:row_stop, col_start:col_stop]

    @property
    def data_array_x_coords(self) -> np.ndarray:
        """
        Get the x coordinates of the data array.

        Returns
        -------
            np.ndarray: The x coordinates of the data array.
        """
        if not isinstance(self._data_array_x_coords, np.ndarray):
            self._data_array_x_coords = self.data_array[self.x_var].to_numpy()
        return self._data_array_x_coords

    @property
    def data_array_y_coords(self) -> np.ndarray:
        """
        Get the y coordinates of the data array.

        Returns
        -------
            np.ndarray: The y coordinates of the data array.
        """
        if not isinstance(self._data_array_y_coords, np.ndarray):
            self._data_array_y_coords = self.data_array[self.y_var].to_numpy()
        return self._data_array_y_coords

    @property
    def np_data_array(self) -> np.ndarray:
        """
        Get the data array as a numpy array.

        Returns
        -------
            np.ndarray: The data array as a numpy array.
        """
        if not isinstance(self._np_data_array, np.ndarray):
            self._np_data_array = self.data_array.to_numpy()
        return self._np_data_array

    @property
    def watershed_window(self) -> Window:
        """
        Get the window of the watershed in the data array.

        Returns
        -------
            Window: The window of the watershed in the data array.
        """
        if self._watershed_window is None:
            self._calculate_watershed_mask_and_window()
        return self._watershed_window

    @property
    def watershed_mask(self) -> np.ndarray:
        """
        Get the watershed mask as a 2D boolean numpy array.

        Returns
        -------
            np.ndarray: The watershed mask as a 2D boolean numpy array.
        """
        if not isinstance(self._watershed_mask, np.ndarray):
            self._calculate_watershed_mask_and_window()
        return self._watershed_mask

    @property
    def watershed_mask_clipped(self) -> np.ndarray:
        """
        Get the clipped watershed mask.

        Returns
        -------
            np.ndarray: The clipped watershed mask.
        """
        if not isinstance(self._watershed_mask_clipped, np.ndarray):
            self._calculate_watershed_mask_and_window()
        return self._watershed_mask_clipped

    @property
    def valid_shifts(self) -> list[tuple[int, int]]:
        """
        Calculate and return a list of valid shift values for the watershed mask.

        This method determines the valid shifts that can be applied to the watershed mask
        within the bounds of the data array. It iterates over possible shifts and checks
        if the shifted mask is still within the valid data region.

        Returns
        -------
            list[tuple[int, int]]: A list of tuples representing the valid (x, y) shifts.
        """
        if self._valid_shifts is None:
            original_window_row_slice, original_window_col_slice = self.watershed_window.toslices()
            shifts: list[tuple[int, int]] = []
            min_x_delta = 0 - self.watershed_window.col_off
            min_y_delta = 0 - self.watershed_window.row_off
            max_x_delta = self.width - (self.watershed_window.col_off + self.watershed_window.width)
            max_y_delta = self.height - (self.watershed_window.row_off + self.watershed_window.height)
            x_delta = min_x_delta
            y_delta = min_y_delta
            while x_delta <= max_x_delta:
                while y_delta <= max_y_delta:
                    adjusted_row_start = original_window_row_slice.start + y_delta
                    adjusted_row_stop = original_window_row_slice.stop + y_delta
                    adjusted_col_start = original_window_col_slice.start + x_delta
                    adjusted_col_stop = original_window_col_slice.stop + x_delta
                    data_clipped = self.np_data_array[
                        adjusted_row_start:adjusted_row_stop, adjusted_col_start:adjusted_col_stop
                    ]
                    data_mask = np.isfinite(data_clipped)
                    combined_mask = np.logical_and(self.watershed_mask_clipped, data_mask)
                    if np.array_equal(combined_mask, self.watershed_mask_clipped):
                        shifts.append((x_delta, y_delta))
                    y_delta += 4
                x_delta += 4
                y_delta = min_y_delta
            self._valid_shifts = shifts

        return self._valid_shifts

    @property
    def valid_spaces(self) -> np.ndarray:
        """
        Calculate and return the valid spaces mask.

        This method initializes the valid mask as the watershed mask value, iterates over
        the list of shift values, applies the shifts to the watershed mask, and performs
        a logical OR with the valid mask and the rolled watershed mask.

        Returns
        -------
            np.ndarray: The valid spaces mask as a boolean numpy array.
        """
        if not isinstance(self._valid_spaces, np.ndarray):
            valid_spaces = np.full(self.watershed_mask.shape, False, dtype=bool)
            for shift in self.valid_shifts:
                rolled = np.roll(self.watershed_mask, shift, axis=(1, 0))
                valid_spaces = np.logical_or(valid_spaces, rolled)
            self._valid_spaces = valid_spaces
        return self._valid_spaces

    def _array_to_polygon(self, arr: np.ndarray) -> Polygon:
        """
        Convert a boolean array to a geometry using the coordinates of the dataset.

        Args:
            arr (np.ndarray): The boolean array to convert.

        Returns
        -------
            Polygon: The resulting polygon.
        """
        shapely_shapes = [
            shape(converted_shape) for converted_shape, _ in shapes(arr.astype(np.ubyte), arr, transform=self.transform)
        ]
        if len(shapely_shapes) != 1:
            raise ValueError(f"Expected single geometry feature, got {len(shapely_shapes)}")
        valid_spaces_geom = shapely_shapes[0]
        if valid_spaces_geom.geom_type != "Polygon":
            raise TypeError(f"Expected geometry type 'Polygon', got {valid_spaces_geom.geom_type}")
        return valid_spaces_geom

    @property
    def valid_spaces_polygon(self) -> Polygon:
        """
        Convert the valid spaces boolean array to a polygon.

        Returns
        -------
            Polygon: The valid spaces polygon.
        """
        if self._valid_spaces_polygon is None:
            self._valid_spaces_polygon = self._array_to_polygon(self.valid_spaces)
        return self._valid_spaces_polygon

    def max_transpose(self, func: Callable[[np.ndarray], Any] | None = None) -> tuple[Polygon, Affine, Any | None]:
        """
        Calculate the maximum transpose of the watershed mask over the data array.

        This method initializes the max transpose array, max shift, and stats collection as None.
        It iterates over the list of shift values, applies the shifts to the watershed mask,
        calculates the stats, and if the stats collection has a greater mean than the max stats,
        it overwrites the max stats, max shift, and max transpose array.

        Args:
            func (Callable[[np.ndarray], Any] | None): A callable to apply to the data array.

        Returns
        -------
            tuple[Polygon, Affine, Any | None]: The resulting polygon, affine transformation, and results.
        """
        original_window_row_slice, original_window_col_slice = self.watershed_window.toslices()
        max_mean = None
        max_shift = None
        results = None
        for x_delta, y_delta in self.valid_shifts:
            adjusted_row_start = original_window_row_slice.start + y_delta
            adjusted_row_stop = original_window_row_slice.stop + y_delta
            adjusted_col_start = original_window_col_slice.start + x_delta
            adjusted_col_stop = original_window_col_slice.stop + x_delta
            data_clipped = self.np_data_array[
                adjusted_row_start:adjusted_row_stop, adjusted_col_start:adjusted_col_stop
            ]
            data_clipped_masked = np.ma.masked_array(data_clipped, ~self.watershed_mask_clipped)
            mean = np.nanmean(data_clipped_masked)
            if max_mean is None or mean > max_mean:
                max_mean = mean
                max_shift = (float(x_delta * self.x_cellsize), float(y_delta * self.y_cellsize))
                if func:
                    results = func(data_clipped_masked)
        poly = self._array_to_polygon(self.watershed_mask)
        poly = translate(poly, *max_shift)
        aff = Affine.translation(*max_shift)
        return poly, aff, results
